#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
import itertools
import json
import logging
from typing import Any

from osc_lib import utils as oscutils

from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc
from ironicclient.osc import command
from ironicclient.v1 import resource_fields as res_fields


_RUNBOOK_STEPS_HELP = _(
    "The runbook steps. May be the path to a YAML file containing the "
    "runbook steps; OR '-', with the runbook steps being read from standard "
    "input; OR a JSON string. The value should be a list of runbook step "
    "dictionaries; each dictionary should have keys 'interface', 'step', "
    "'args' and 'order'.")


class CreateBaremetalRunbook(command.ShowOne):
    """Create a new runbook"""

    log: logging.Logger = logging.getLogger(
        __name__ + ".CreateBaremetalRunbook")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--name',
            metavar='<name>',
            required=True,
            help=_('Unique name for this runbook.')
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Description of the runbook.')
        )
        parser.add_argument(
            '--traits',
            metavar='<trait>',
            action='append',
            help=_('Trait for this runbook. Can be specified multiple times.')
        )
        parser.add_argument(
            '--uuid',
            dest='uuid',
            metavar='<uuid>',
            help=_('UUID of the runbook.'))
        parser.add_argument(
            '--public',
            dest='public',
            nargs='?',
            const='true',
            metavar='<public>',
            help=_('Whether the runbook will be private or public.')
        )
        parser.add_argument(
            '--owner',
            metavar='<owner>',
            help=_('Owner of the runbook.')
        )
        parser.add_argument(
            '--extra',
            metavar="<key=value>",
            action='append',
            help=_("Record arbitrary key/value metadata. "
                   "Can be specified multiple times."))
        parser.add_argument(
            '--steps',
            metavar="<steps>",
            required=True,
            help=_RUNBOOK_STEPS_HELP
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        steps = utils.handle_json_arg(parsed_args.steps, 'runbook steps')

        field_list: list[str] = ['name', 'uuid', 'owner', 'public', 'extra']

        # Check if API version supports new fields (description and traits)
        if utils.check_api_version_support(
                baremetal_client.current_api_version, "1.112"):
            field_list.extend(['description', 'traits'])
        else:
            if parsed_args.description is not None:
                raise exc.UnsupportedVersion(
                    _("--description requires API version 1.112 or later"))
            if parsed_args.traits is not None:
                raise exc.UnsupportedVersion(
                    _("--traits requires API version 1.112 or later"))

        fields = dict((k, v) for (k, v) in vars(parsed_args).items()
                      if k in field_list and v is not None)
        if parsed_args.public is not None:
            fields['public'] = utils.bool_argument_value(
                '--public', parsed_args.public, strict=True)
        fields = utils.args_array_to_dict(fields, 'extra')
        runbook = baremetal_client.runbook.create(steps=steps,
                                                  **fields)

        data = dict([(f, getattr(runbook, f, '')) for f in
                     res_fields.RUNBOOK_DETAILED_RESOURCE.fields])

        return self.dict2columns(data)


class ShowBaremetalRunbook(command.ShowOne):
    """Show baremetal runbook details."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ShowBaremetalRunbook")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "runbook",
            metavar="<runbook>",
            help=_("Name or UUID of the runbook.")
        )
        parser.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            choices=res_fields.RUNBOOK_DETAILED_RESOURCE.fields,
            default=[],
            help=_("One or more runbook fields. Only these fields "
                   "will be fetched from the server.")
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal
        fields: list[str] | None
        fields = (
            list(itertools.chain.from_iterable(parsed_args.fields))
            or None)

        runbook = baremetal_client.runbook.get(
            parsed_args.runbook, fields=fields)._info

        runbook.pop("links", None)
        return self.dict2columns(dict(sorted(runbook.items())))


class SetBaremetalRunbook(command.Command):
    """Set baremetal runbook properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".SetBaremetalRunbook")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'runbook',
            metavar='<runbook>',
            help=_("Name or UUID of the runbook")
        )
        parser.add_argument(
            '--name',
            metavar='<name>',
            help=_('Set unique name of the runbook.')
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Set description of the runbook.')
        )
        parser.add_argument(
            '--public',
            dest='public',
            nargs='?',
            const='true',
            metavar='<public>',
            help=_('Make a private runbook public.')
        )
        parser.add_argument(
            '--owner',
            metavar='<owner>',
            help=_('Set owner of a runbook.')
        )
        parser.add_argument(
            '--steps',
            metavar="<steps>",
            help=_RUNBOOK_STEPS_HELP
        )
        parser.add_argument(
            "--extra",
            metavar="<key=value>",
            action='append',
            help=_('Extra to set on this baremetal runbook '
                   '(repeat option to set multiple extras).'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        # Check if API version supports description field
        supports_new_fields = utils.check_api_version_support(
            baremetal_client.current_api_version, "1.112")

        if parsed_args.name:
            name = ["name=%s" % parsed_args.name]
            properties.extend(utils.args_array_to_patch('add', name))
        if parsed_args.description:
            if not supports_new_fields:
                raise exc.UnsupportedVersion(
                    _("--description requires API version 1.112 or later"))
            description = ["description=%s" % parsed_args.description]
            properties.extend(utils.args_array_to_patch('add', description))
        if parsed_args.owner:
            owner = ["owner=%s" % parsed_args.owner]
            properties.extend(utils.args_array_to_patch('add', owner))
        if parsed_args.public is not None:
            to_bool = utils.bool_argument_value(
                '--public', parsed_args.public, strict=True)
            public = ["public=%s" % to_bool]
            properties.extend(utils.args_array_to_patch('add', public))
        if parsed_args.steps:
            steps = utils.handle_json_arg(parsed_args.steps, 'runbook steps')
            steps = ["steps=%s" % json.dumps(steps)]
            properties.extend(utils.args_array_to_patch('add', steps))
        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch(
                'add', ['extra/' + x for x in parsed_args.extra]))

        if properties:
            baremetal_client.runbook.update(parsed_args.runbook,
                                            properties)
        else:
            self.log.warning("Please specify what to set.")


class UnsetBaremetalRunbook(command.Command):
    """Unset baremetal runbook properties."""
    log: logging.Logger = logging.getLogger(
        __name__ + ".UnsetBaremetalRunbook")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'runbook',
            metavar='<runbook>',
            help=_("Name or UUID of the runbook")
        )
        parser.add_argument(
            '--name',
            action='store_true',
            help=_('Unset the name of the runbook.')
        )
        parser.add_argument(
            '--description',
            action='store_true',
            help=_('Unset the description of the runbook.')
        )
        parser.add_argument(
            '--public',
            dest='public',
            action='store_true',
            help=_('Make a public runbook private.')
        )
        parser.add_argument(
            '--owner',
            dest='owner',
            action='store_true',
            help=_('Unset owner of a runbook.')
        )
        parser.add_argument(
            "--step",
            metavar="<key>",
            action='append',
            help=_('Step to unset on this baremetal runbook '
                   '(repeat option to unset multiple steps).'),
        )
        parser.add_argument(
            "--extra",
            metavar="<key>",
            action='append',
            help=_('Extra to unset on this baremetal runbook '
                   '(repeat option to unset multiple extras).'),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        # Check if API version supports description field
        supports_new_fields = utils.check_api_version_support(
            baremetal_client.current_api_version, "1.112")

        for field in ['name', 'owner', 'public']:
            if getattr(parsed_args, field):
                properties.extend(utils.args_array_to_patch('remove', [field]))

        # Handle description field separately with version check
        if getattr(parsed_args, 'description', False):
            if not supports_new_fields:
                raise exc.UnsupportedVersion(
                    _("--description requires API version 1.112 or later"))
            properties.extend(
                utils.args_array_to_patch('remove', ['description']))

        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch('remove',
                              ['extra/' + x for x in parsed_args.extra]))
        if parsed_args.step:
            properties.extend(utils.args_array_to_patch('remove',
                              ['step/' + x for x in parsed_args.step]))

        if properties:
            baremetal_client.runbook.update(parsed_args.runbook,
                                            properties)
        else:
            self.log.warning("Please specify what to unset.")


class DeleteBaremetalRunbook(command.Command):
    """Delete runbook(s)."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".DeleteBaremetalRunbook")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "runbooks",
            metavar="<runbook>",
            nargs="+",
            help=_("Name(s) or UUID(s) of the runbook(s) to delete.")
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        failures: list[str] = []
        for runbook in parsed_args.runbooks:
            try:
                baremetal_client.runbook.delete(runbook)
                print(_('Deleted runbook %s') % runbook)
            except exc.ClientException as e:
                failures.append(_("Failed to delete runbook "
                                  "%(runbook)s: %(error)s")
                                % {'runbook': runbook, 'error': e})

        if failures:
            raise exc.ClientException("\n".join(failures))


class ListBaremetalRunbook(command.Lister):
    """List baremetal runbooks."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalRunbook")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of runbooks to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Baremetal API Service.')
        )
        parser.add_argument(
            '--marker',
            metavar='<runbook>',
            help=_('Runbook UUID (for example, of the last runbook '
                   'in the list from a previous request). Returns '
                   'the list of runbooks after this UUID.')
        )
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            help=_('Sort output by specified runbook fields and '
                   'directions (asc or desc) (default: asc). Multiple fields '
                   'and directions can be specified, separated by comma.')
        )
        display_group = parser.add_mutually_exclusive_group()
        display_group.add_argument(
            '--long',
            dest='detail',
            action='store_true',
            default=False,
            help=_("Show detailed information about runbooks.")
        )
        display_group.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.RUNBOOK_DETAILED_RESOURCE.fields,
            help=_("One or more runbook fields. Only these fields "
                   "will be fetched from the server. Can not be used when "
                   "'--long' is specified.")
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        client = manager.baremetal

        columns = res_fields.RUNBOOK_RESOURCE.fields

        params: dict[str, object] = {}
        if parsed_args.limit is not None and parsed_args.limit < 0:
            raise exc.CommandError(
                _('Expected non-negative --limit, got %s') %
                parsed_args.limit)
        params['limit'] = parsed_args.limit
        params['marker'] = parsed_args.marker

        if parsed_args.detail:
            params['detail'] = parsed_args.detail
            columns = res_fields.RUNBOOK_DETAILED_RESOURCE.fields

        elif parsed_args.fields:
            params['detail'] = False
            fields = itertools.chain.from_iterable(parsed_args.fields)
            resource = res_fields.Resource(list(fields))
            columns = resource.fields
            params['fields'] = columns

        self.log.debug("params(%s)", params)
        data = client.runbook.list(**params)

        data = oscutils.sort_items(data, parsed_args.sort)

        return (columns,
                (oscutils.get_item_properties(s, columns) for s in data))


class ListTraitsBaremetalRunbook(command.Lister):
    """List a runbook's traits."""

    log = logging.getLogger(__name__ + ".ListTraitsBaremetalRunbook")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super(ListTraitsBaremetalRunbook, self).get_parser(prog_name)

        parser.add_argument(
            'runbook',
            metavar='<runbook>',
            help=_("Name or UUID of the runbook"))

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], list[list[str]]]:
        self.log.debug("take_action(%s)", parsed_args)

        columns = res_fields.TRAIT_RESOURCE.fields

        baremetal_client = self.app.client_manager.baremetal
        traits = baremetal_client.runbook.get_traits(parsed_args.runbook)

        return (columns, [[trait] for trait in traits])


class AddTraitBaremetalRunbook(command.Command):
    """Add traits to a runbook."""

    log = logging.getLogger(__name__ + ".AddTraitBaremetalRunbook")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super(AddTraitBaremetalRunbook, self).get_parser(prog_name)

        parser.add_argument(
            'runbook',
            metavar='<runbook>',
            help=_("Name or UUID of the runbook"))
        parser.add_argument(
            'traits',
            nargs='+',
            metavar='<trait>',
            help=_("Trait(s) to add"))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        baremetal_client = self.app.client_manager.baremetal

        failures = []
        for trait in parsed_args.traits:
            try:
                baremetal_client.runbook.add_trait(parsed_args.runbook, trait)
                print(_('Added trait %s') % trait)
            except exc.ClientException as e:
                failures.append(_("Failed to add trait %(trait)s: %(error)s")
                                % {'trait': trait, 'error': e})

        if failures:
            raise exc.ClientException("\n".join(failures))


class RemoveTraitBaremetalRunbook(command.Command):
    """Remove trait(s) from a runbook."""

    log = logging.getLogger(__name__ + ".RemoveTraitBaremetalRunbook")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super(RemoveTraitBaremetalRunbook, self).get_parser(prog_name)

        parser.add_argument(
            'runbook',
            metavar='<runbook>',
            help=_("Name or UUID of the runbook"))
        all_or_trait = parser.add_mutually_exclusive_group(required=True)
        all_or_trait.add_argument(
            '--all',
            dest='remove_all',
            action='store_true',
            help=_("Remove all traits"))
        all_or_trait.add_argument(
            'traits',
            metavar='<trait>',
            nargs='*',
            default=[],
            help=_("Trait(s) to remove"))

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        baremetal_client = self.app.client_manager.baremetal

        failures = []
        if parsed_args.remove_all:
            baremetal_client.runbook.remove_all_traits(parsed_args.runbook)
            print(_('Removed all traits from runbook %s')
                  % parsed_args.runbook)
        else:
            for trait in parsed_args.traits:
                try:
                    baremetal_client.runbook.remove_trait(
                        parsed_args.runbook, trait)
                    print(_('Removed trait %s') % trait)
                except exc.ClientException as e:
                    failures.append(_("Failed to remove trait %(trait)s: "
                                      "%(error)s")
                                    % {'trait': trait, 'error': e})

        if failures:
            raise exc.ClientException("\n".join(failures))



#
#   Copyright 2015 Red Hat, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""OpenStackClient plugin for Bare Metal service."""

from __future__ import annotations

import argparse
import logging

from osc_lib import clientmanager
from osc_lib import utils

from ironicclient.common import http
from ironicclient.v1 import client as v1_client

LOG: logging.Logger = logging.getLogger(__name__)

CLIENT_CLASS: str = 'ironicclient.v1.client.Client'
API_VERSION_OPTION: str = 'os_baremetal_api_version'
API_NAME: str = 'baremetal'
# NOTE(TheJulia) Latest known version tracking has been moved
# to the ironicclient/common/http.py file as the OSC commitment
# is latest known, and we should only store it in one location.
LAST_KNOWN_API_VERSION: int = http.LAST_KNOWN_API_VERSION
LATEST_VERSION: str = http.LATEST_VERSION


API_VERSIONS: dict[str, str] = {
    '1.%d' % i: CLIENT_CLASS
    for i in range(1, LAST_KNOWN_API_VERSION + 1)
}
API_VERSIONS['1'] = CLIENT_CLASS
# NOTE(dtantsur): flag to indicate that the requested version was "latest".
# Due to how OSC works we cannot just add "latest" to the list of supported
# versions - it breaks the major version detection.
OS_BAREMETAL_API_LATEST: bool = True


def make_client(
    instance: clientmanager.ClientManager,
) -> v1_client.Client:
    """Returns a baremetal service client."""
    requested_api_version: str = instance._api_version[API_NAME]

    baremetal_client_class: type[v1_client.Client] = (
        utils.get_client_class(
            API_NAME,
            requested_api_version,
            API_VERSIONS))
    LOG.debug('Instantiating baremetal client: %s', baremetal_client_class)
    LOG.debug('Baremetal API version: %s',
              requested_api_version if not OS_BAREMETAL_API_LATEST
              else "latest")

    if requested_api_version == '1':
        # NOTE(dtantsur): '1' means 'the latest v1 API version'. Since we don't
        # have other major versions, it's identical to 'latest'.
        requested_api_version = LATEST_VERSION
        allow_api_version_downgrade: bool = True
    else:
        allow_api_version_downgrade = OS_BAREMETAL_API_LATEST

    result: v1_client.Client = baremetal_client_class(
        os_ironic_api_version=requested_api_version,
        # NOTE(dtantsur): enable re-negotiation of the latest version, if CLI
        # latest is too high for the server we're talking to.
        allow_api_version_downgrade=allow_api_version_downgrade,
        session=instance.session,
        # TODO(anandkaranubc): _region_name is a private attr on ClientManager;
        # osc-lib should expose this publicly. See also:
        # https://bugs.launchpad.net/python-ironicclient/+bug/2146739
        region_name=instance._region_name,  # type: ignore[attr-defined]
        # NOTE(vdrok): This will be set as endpoint_override, and the Client
        # class will be able to do the version stripping if needed
        endpoint_override=instance.get_endpoint_for_service_type(
            API_NAME, interface=instance.interface,
            # TODO(anandkaranubc): same _region_name private
            # attr access as above
            region_name=instance._region_name  # type: ignore[attr-defined]
        )
    )
    return result


def build_option_parser(
    parser: argparse.ArgumentParser,
) -> argparse.ArgumentParser:
    """Hook to add global options."""
    parser.add_argument(
        '--os-baremetal-api-version',
        metavar='<baremetal-api-version>',
        default=_get_environment_version("latest"),
        choices=sorted(
            API_VERSIONS,
            key=lambda k: [int(x) for x in k.split('.')]) + ['latest'],
        action=ReplaceLatestVersion,
        help='Bare metal API version, default="latest" (the maximum version '
             'supported by both the client and the server). '
             '(Env: OS_BAREMETAL_API_VERSION)',
    )
    return parser


def _get_environment_version(default: str) -> str:
    global OS_BAREMETAL_API_LATEST
    env_value: str | None = utils.env('OS_BAREMETAL_API_VERSION')
    if not env_value:
        env_value = default
    if env_value == 'latest':
        env_value = LATEST_VERSION
    else:
        OS_BAREMETAL_API_LATEST = False
    return env_value


class ReplaceLatestVersion(argparse.Action):
    """Replaces `latest` keyword by last known version.

    OSC cannot accept the literal "latest" as a supported API version as it
    breaks the major version detection (OSC tries to load configuration options
    from setuptools entrypoint openstack.baremetal.vlatest). This action
    replaces "latest" with the latest known version, and sets the global
    OS_BAREMETAL_API_LATEST flag appropriately.
    """
    # base Action.__call__ accepts str | Sequence[Any] | None; narrowed here
    def __call__(  # type: ignore[override]
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str,
        option_string: str | None = None,
    ) -> None:
        global OS_BAREMETAL_API_LATEST
        if values == 'latest':
            values = LATEST_VERSION
            # The default value of "True" may have been overridden due to
            # non-empty OS_BAREMETAL_API_VERSION env variable. If a user
            # explicitly requests "latest", we need to correct it.
            OS_BAREMETAL_API_LATEST = True
        else:
            OS_BAREMETAL_API_LATEST = False
        setattr(namespace, self.dest, values)



# -- General configuration ----------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom ones.
extensions = ['sphinxcontrib.apidoc',
              'sphinx.ext.viewcode',
              'openstackdocstheme',
              'cliff.sphinxext',
              ]

# sphinxcontrib.apidoc options
apidoc_module_dir = '../../ironicclient'
apidoc_output_dir = 'reference/api'
apidoc_excluded_paths = [
    'tests']
apidoc_separate_modules = True


# openstackdocstheme options
openstackdocs_repo_name = 'openstack/python-ironicclient'
openstackdocs_bug_project = 'python-ironicclient'
openstackdocs_pdf_link = True

# autodoc generation is a bit aggressive and a nuisance when doing heavy
# text edit cycles.
# execute "export SPHINX_DEBUG=1" in your terminal to disable

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix of source filenames.
source_suffix = '.rst'

# The master toctree document.
master_doc = 'index'

# General information about the project.
copyright = 'OpenStack Foundation'

# A list of ignored prefixes for module index sorting.
modindex_common_prefix = ['ironicclient.']

# If true, '()' will be appended to :func: etc. cross-reference text.
add_function_parentheses = True

# If true, the current module name will be prepended to all description
# unit titles (such as .. function::).
add_module_names = True

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'native'

# A list of glob-style patterns that should be excluded when looking for
# source files. They are matched against the source file names relative to the
# source directory, using slashes as directory separators on all platforms.
exclude_patterns = ['']

# -- Options for HTML output --------------------------------------------------

# The theme to use for HTML and HTML Help pages. Major themes that come with
# Sphinx are currently 'default' and 'sphinxdoc'.
#html_theme_path = ["."]
#html_theme = '_theme'
#html_static_path = ['_static']
html_theme = 'openstackdocs'

# Output file base name for HTML help builder.
htmlhelp_basename = 'python-ironicclientdoc'

latex_use_xindy = False

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title, author, documentclass
# [howto/manual]).
latex_documents = [
    (
        'index',
        'doc-python-ironicclient.tex',
        'Python Ironic Client Documentation',
        'OpenStack Ironic Project',
        'manual'
    ),
]

autoprogram_cliff_application = 'openstack'



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from unittest import mock

from ironicclient.common import filecache
from ironicclient.common import http
from ironicclient.tests.unit import utils
from ironicclient.v1 import client


@mock.patch.object(http, '_construct_http_client', autospec=True)
class ClientTest(utils.BaseTestCase):

    session = mock.sentinel.session

    def test_client_user_api_version(
        self, http_client_mock: mock.MagicMock,
    ) -> None:
        endpoint = 'http://ironic:6385'
        os_ironic_api_version = '1.15'

        client.Client(endpoint, session=self.session,
                      os_ironic_api_version=os_ironic_api_version)

        http_client_mock.assert_called_once_with(
            endpoint_override=endpoint, session=self.session,
            os_ironic_api_version=os_ironic_api_version,
            api_version_select_state='user')

    def test_client_user_api_version_with_downgrade(
        self, http_client_mock: mock.MagicMock,
    ) -> None:
        endpoint = 'http://ironic:6385'
        os_ironic_api_version = '1.15'

        client.Client(endpoint, session=self.session,
                      os_ironic_api_version=os_ironic_api_version,
                      allow_api_version_downgrade=True)

        http_client_mock.assert_called_once_with(
            endpoint_override=endpoint, session=self.session,
            os_ironic_api_version=os_ironic_api_version,
            api_version_select_state='default')

    def test_client_user_api_version_latest_with_downgrade(
        self,
        http_client_mock: mock.MagicMock,
    ) -> None:
        endpoint = 'http://ironic:6385'
        os_ironic_api_version = 'latest'

        self.assertRaises(ValueError, client.Client, endpoint,
                          session=self.session,
                          allow_api_version_downgrade=True,
                          os_ironic_api_version=os_ironic_api_version)

    @mock.patch.object(filecache, 'retrieve_data', autospec=True)
    def test_client_cache_api_version(
        self,
        cache_mock: mock.MagicMock,
        http_client_mock: mock.MagicMock,
    ) -> None:
        endpoint = 'http://ironic:6385'
        os_ironic_api_version = '1.15'
        cache_mock.return_value = os_ironic_api_version

        client.Client(endpoint, session=self.session)

        cache_mock.assert_called_once_with(host='ironic', port='6385')
        http_client_mock.assert_called_once_with(
            endpoint_override=endpoint, session=self.session,
            os_ironic_api_version=os_ironic_api_version,
            api_version_select_state='cached')

    @mock.patch.object(filecache, 'retrieve_data', autospec=True)
    def test_client_default_api_version(
        self,
        cache_mock: mock.MagicMock,
        http_client_mock: mock.MagicMock,
    ) -> None:
        endpoint = 'http://ironic:6385'
        cache_mock.return_value = None

        client.Client(endpoint, session=self.session)

        cache_mock.assert_called_once_with(host='ironic', port='6385')
        http_client_mock.assert_called_once_with(
            endpoint_override=endpoint, session=self.session,
            os_ironic_api_version=client.DEFAULT_VER,
            api_version_select_state='default')

    def test_client_cache_version_no_endpoint_as_arg(
        self, http_client_mock: mock.MagicMock,
    ) -> None:
        client.Client(session='fake_session', insecure=True)
        http_client_mock.assert_called_once_with(
            session='fake_session', insecure=True,
            os_ironic_api_version=client.DEFAULT_VER,
            api_version_select_state='default')

    def test_client_initialized_managers(
        self, http_client_mock: mock.MagicMock,
    ) -> None:
        cl = client.Client('http://ironic:6385', session=self.session,
                           os_ironic_api_version='1.15')

        self.assertIsInstance(cl.node, client.node.NodeManager)
        self.assertIsInstance(cl.port, client.port.PortManager)
        self.assertIsInstance(cl.driver, client.driver.DriverManager)
        self.assertIsInstance(cl.chassis, client.chassis.ChassisManager)

    def test_negotiate_api_version(
        self, http_client_mock: mock.MagicMock,
    ) -> None:
        endpoint = 'http://ironic:6385'
        os_ironic_api_version = 'latest'
        cl = client.Client(endpoint, session=self.session,
                           os_ironic_api_version=os_ironic_api_version)

        cl.negotiate_api_version()
        http_client_mock.assert_called_once_with(
            api_version_select_state='user', endpoint_override=endpoint,
            os_ironic_api_version='latest', session=self.session)
        # TODO(TheJulia): We should verify that negotiate_version
        # is being called in the client and returns a version,
        # although mocking might need to be restrutured to
        # properly achieve that.

    def test_client_no_session(
        self, http_client_mock: mock.MagicMock,
    ) -> None:
        self.assertRaisesRegex(TypeError,
                               "session is required",
                               client.Client,
                               "http://example.com")

    def test_client_session_via_posargs(
        self, http_client_mock: mock.MagicMock,
    ) -> None:
        session = mock.Mock()
        client.Client("http://example.com", session)
        http_client_mock.assert_called_once_with(
            session, api_version_select_state='default',
            endpoint_override="http://example.com",
            os_ironic_api_version=client.DEFAULT_VER)

    def test_client_session_via_kwargs(
        self, http_client_mock: mock.MagicMock,
    ) -> None:
        session = mock.Mock()
        client.Client(session=session, endpoint_override="http://example.com")
        http_client_mock.assert_called_once_with(
            session, api_version_select_state='default',
            endpoint_override="http://example.com",
            os_ironic_api_version=client.DEFAULT_VER)



# Copyright 2015 Hitachi Data Systems
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import copy

import testtools

from ironicclient import exc
from ironicclient.tests.unit import utils
import ironicclient.v1.port
import ironicclient.v1.volume_connector

NODE_UUID = '55555555-4444-3333-2222-111111111111'
CONNECTOR1 = {'uuid': '11111111-2222-3333-4444-555555555555',
              'node_uuid': NODE_UUID,
              'type': 'iqn',
              'connector_id': 'iqn.2010-10.org.openstack:test',
              'extra': {}}

CONNECTOR2 = {'uuid': '66666666-7777-8888-9999-000000000000',
              'node_uuid': NODE_UUID,
              'type': 'wwpn',
              'connector_id': '1234567890543216',
              'extra': {}}

CREATE_CONNECTOR = copy.deepcopy(CONNECTOR1)
del CREATE_CONNECTOR['uuid']

CREATE_CONNECTOR_WITH_UUID = copy.deepcopy(CONNECTOR1)

UPDATED_CONNECTOR = copy.deepcopy(CONNECTOR1)
NEW_CONNECTOR_ID = '1234567890123456'
UPDATED_CONNECTOR['connector_id'] = NEW_CONNECTOR_ID

fake_responses = {
    '/v1/volume/connectors':
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR1]},
        ),
        'POST': (
            {},
            CONNECTOR1,
        ),
    },
    '/v1/volume/connectors/?detail=True':
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR1]},
        ),
    },
    '/v1/volume/connectors/?fields=uuid,connector_id':
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR1]},
        ),
    },
    '/v1/volume/connectors/%s' % CONNECTOR1['uuid']:
    {
        'GET': (
            {},
            CONNECTOR1,
        ),
        'DELETE': (
            {},
            None,
        ),
        'PATCH': (
            {},
            UPDATED_CONNECTOR,
        ),
    },
    '/v1/volume/connectors/%s?fields=uuid,connector_id' % CONNECTOR1['uuid']:
    {
        'GET': (
            {},
            CONNECTOR1,
        ),
    },
    '/v1/volume/connectors/?detail=True&node=%s' % NODE_UUID:
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR1]},
        ),
    },
    '/v1/volume/connectors/?node=%s' % NODE_UUID:
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR1]},
        ),
    }
}

fake_responses_pagination = {
    '/v1/volume/connectors':
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR1],
             "next": "http://127.0.0.1:6385/v1/volume/connectors/?marker=%s" %
                     CONNECTOR1['uuid']}
        ),
    },
    '/v1/volume/connectors/?limit=1':
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR1],
             "next": "http://127.0.0.1:6385/v1/volume/connectors/?limit=1"
                     "&marker=%s" % CONNECTOR1['uuid']}
        ),
    },
    '/v1/volume/connectors/?limit=1&marker=%s' % CONNECTOR1['uuid']:
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR2]}
        ),
    },
    '/v1/volume/connectors/?marker=%s' % CONNECTOR1['uuid']:
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/volume/connectors/?sort_key=updated_at':
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR2, CONNECTOR1]}
        ),
    },
    '/v1/volume/connectors/?sort_dir=desc':
    {
        'GET': (
            {},
            {"connectors": [CONNECTOR2, CONNECTOR1]}
        ),
    },
}


class VolumeConnectorManagerTestBase(testtools.TestCase):

    def _validate_obj(self, expect: dict[str, object],
                      obj: object) -> None:
        self.assertEqual(expect['uuid'], obj.uuid)
        self.assertEqual(expect['type'], obj.type)
        self.assertEqual(expect['connector_id'], obj.connector_id)
        self.assertEqual(expect['node_uuid'], obj.node_uuid)

    def _validate_list(self, expect_request: list[object],
                       expect_connectors: list[dict[str, object]],
                       actual_connectors: list[object]) -> None:
        self.assertEqual(expect_request, self.api.calls)
        self.assertEqual(len(expect_connectors), len(actual_connectors))
        for expect, obj in zip(expect_connectors, actual_connectors):
            self._validate_obj(expect, obj)


class VolumeConnectorManagerTest(VolumeConnectorManagerTestBase):

    def setUp(self) -> None:
        super(VolumeConnectorManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = ironicclient.v1.volume_connector.VolumeConnectorManager(
            self.api)

    def test_volume_connectors_list(self) -> None:
        volume_connectors = self.mgr.list()
        expect = [
            ('GET', '/v1/volume/connectors', {}, None),
        ]
        expect_connectors = [CONNECTOR1]
        self._validate_list(expect, expect_connectors, volume_connectors)

    def test_volume_connectors_list_by_node(self) -> None:
        volume_connectors = self.mgr.list(node=NODE_UUID)
        expect = [
            ('GET', '/v1/volume/connectors/?node=%s' % NODE_UUID, {}, None),
        ]
        expect_connectors = [CONNECTOR1]
        self._validate_list(expect, expect_connectors, volume_connectors)

    def test_volume_connectors_list_by_node_detail(self) -> None:
        volume_connectors = self.mgr.list(node=NODE_UUID, detail=True)
        expect = [
            ('GET', '/v1/volume/connectors/?detail=True&node=%s' % NODE_UUID,
             {}, None),
        ]
        expect_connectors = [CONNECTOR1]
        self._validate_list(expect, expect_connectors, volume_connectors)

    def test_volume_connectors_list_detail(self) -> None:
        volume_connectors = self.mgr.list(detail=True)
        expect = [
            ('GET', '/v1/volume/connectors/?detail=True', {}, None),
        ]
        expect_connectors = [CONNECTOR1]
        self._validate_list(expect, expect_connectors, volume_connectors)

    def test_volume_connector_list_fields(self) -> None:
        volume_connectors = self.mgr.list(fields=['uuid', 'connector_id'])
        expect = [
            ('GET',
             '/v1/volume/connectors/?fields=uuid,connector_id',
             {},
             None),
        ]
        expect_connectors = [CONNECTOR1]
        self._validate_list(expect, expect_connectors, volume_connectors)

    def test_volume_connector_list_detail_and_fields_fail(self) -> None:
        self.assertRaises(exc.InvalidAttribute, self.mgr.list,
                          detail=True, fields=['uuid', 'connector_id'])

    def test_volume_connectors_show(self) -> None:
        volume_connector = self.mgr.get(CONNECTOR1['uuid'])
        expect = [
            ('GET', '/v1/volume/connectors/%s' % CONNECTOR1['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self._validate_obj(CONNECTOR1, volume_connector)

    def test_volume_connector_show_fields(self) -> None:
        volume_connector = self.mgr.get(CONNECTOR1['uuid'],
                                        fields=['uuid', 'connector_id'])
        expect = [
            ('GET', '/v1/volume/connectors/%s?fields=uuid,connector_id' %
             CONNECTOR1['uuid'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(CONNECTOR1['uuid'], volume_connector.uuid)
        self.assertEqual(CONNECTOR1['connector_id'],
                         volume_connector.connector_id)

    def test_create(self) -> None:
        volume_connector = self.mgr.create(**CREATE_CONNECTOR)
        expect = [
            ('POST', '/v1/volume/connectors', {}, CREATE_CONNECTOR),
        ]
        self.assertEqual(expect, self.api.calls)
        self._validate_obj(CONNECTOR1, volume_connector)

    def test_create_with_uuid(self) -> None:
        volume_connector = self.mgr.create(
            **CREATE_CONNECTOR_WITH_UUID)
        expect = [
            ('POST', '/v1/volume/connectors', {},
             CREATE_CONNECTOR_WITH_UUID),
        ]
        self.assertEqual(expect, self.api.calls)
        self._validate_obj(CREATE_CONNECTOR_WITH_UUID,
                           volume_connector)

    def test_delete(self) -> None:
        self.mgr.delete(CONNECTOR1['uuid'])
        expect = [
            ('DELETE',
             '/v1/volume/connectors/%s' % CONNECTOR1['uuid'],
             {}, None),
        ]
        self.assertEqual(expect, self.api.calls)

    def test_update(self) -> None:
        patch = {'op': 'replace',
                 'connector_id': NEW_CONNECTOR_ID,
                 'path': '/connector_id'}
        volume_connector = self.mgr.update(
            volume_connector_id=CONNECTOR1['uuid'],
            patch=patch)
        expect = [
            ('PATCH',
             '/v1/volume/connectors/%s' % CONNECTOR1['uuid'],
             {}, patch),
        ]
        self.assertEqual(expect, self.api.calls)
        self._validate_obj(UPDATED_CONNECTOR, volume_connector)


class VolumeConnectorManagerPaginationTest(VolumeConnectorManagerTestBase):

    def setUp(self) -> None:
        super(VolumeConnectorManagerPaginationTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = ironicclient.v1.volume_connector.VolumeConnectorManager(
            self.api)

    def test_volume_connectors_list_limit(self) -> None:
        volume_connectors = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/volume/connectors/?limit=1', {}, None),
        ]
        expect_connectors = [CONNECTOR1]
        self._validate_list(expect, expect_connectors, volume_connectors)

    def test_volume_connectors_list_marker(self) -> None:
        volume_connectors = self.mgr.list(
            marker=CONNECTOR1['uuid'])
        expect = [
            ('GET', '/v1/volume/connectors/?marker=%s' % CONNECTOR1['uuid'],
             {}, None),
        ]
        expect_connectors = [CONNECTOR2]
        self._validate_list(expect, expect_connectors, volume_connectors)

    def test_volume_connectors_list_pagination_no_limit(self) -> None:
        volume_connectors = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/volume/connectors', {}, None),
            ('GET', '/v1/volume/connectors/?marker=%s' % CONNECTOR1['uuid'],
             {}, None)
        ]
        expect_connectors = [CONNECTOR1, CONNECTOR2]
        self._validate_list(expect, expect_connectors, volume_connectors)


class VolumeConnectorManagerSortingTest(VolumeConnectorManagerTestBase):

    def setUp(self) -> None:
        super(VolumeConnectorManagerSortingTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = ironicclient.v1.volume_connector.VolumeConnectorManager(
            self.api)

    def test_volume_connectors_list_sort_key(self) -> None:
        volume_connectors = self.mgr.list(sort_key='updated_at')
        expect = [
            ('GET', '/v1/volume/connectors/?sort_key=updated_at', {}, None)
        ]
        expect_connectors = [CONNECTOR2, CONNECTOR1]
        self._validate_list(expect, expect_connectors, volume_connectors)

    def test_volume_connectors_list_sort_dir(self) -> None:
        volume_connectors = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/volume/connectors/?sort_dir=desc', {}, None)
        ]
        expect_connectors = [CONNECTOR2, CONNECTOR1]
        self._validate_list(expect, expect_connectors, volume_connectors)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
import itertools
import json
import logging
from typing import Any

from osc_lib import utils as oscutils

from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc
from ironicclient.osc import command
from ironicclient.v1 import resource_fields as res_fields


class CreateBaremetalInspectionRule(command.ShowOne):
    """Create a new rule"""

    log: logging.Logger = logging.getLogger(
        __name__ + ".CreateBaremetalInspectionRule")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--uuid',
            dest='uuid',
            metavar='<uuid>',
            help=_('UUID of the rule.'))
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('A brief explanation about the rule.')
        )
        parser.add_argument(
            '--priority',
            metavar='<priority>',
            help=_("Specifies the rule's precedence level during execution.")
        )
        parser.add_argument(
            '--sensitive',
            metavar='<sensitive>',
            help=_('Indicates whether the rule contains sensitive '
                   'information.')
        )
        parser.add_argument(
            '--phase',
            metavar='<phase>',
            help=_('Specifies the processing phase when the rule should run.')
        )
        parser.add_argument(
            '--conditions',
            metavar="<conditions>",
            help=_('Conditions under which the rule should be triggered.')
        )
        parser.add_argument(
            '--actions',
            metavar="<actions>",
            required=True,
            help=_('Actions to be executed when the rule conditions are met.')
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        actions = utils.handle_json_arg(parsed_args.actions, 'rule actions')
        conditions = utils.handle_json_arg(parsed_args.conditions,
                                           'rule conditions')

        field_list = ['uuid', 'description', 'priority', 'sensitive', 'phase']
        fields = dict((k, v) for (k, v) in vars(parsed_args).items()
                      if k in field_list and v is not None)
        rule = baremetal_client.inspection_rule.create(actions=actions,
                                                       conditions=conditions,
                                                       **fields)

        data = dict([(f, getattr(rule, f, '')) for f in
                     res_fields.INSPECTION_RULE_DETAILED_RESOURCE.fields])

        return self.dict2columns(data)


class ShowBaremetalInspectionRule(command.ShowOne):
    """Show baremetal rule details."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ShowBaremetalInspectionRule")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "rule",
            metavar="<rule>",
            help=_("UUID of the inspection rule.")
        )
        parser.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            choices=res_fields.INSPECTION_RULE_DETAILED_RESOURCE.fields,
            default=[],
            help=_("One or more rule fields. Only these fields "
                   "will be fetched from the server.")
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal
        fields: list[str] | None
        fields = (
            list(itertools.chain.from_iterable(parsed_args.fields))
            or None)

        rule = baremetal_client.inspection_rule.get(
            parsed_args.rule, fields=fields)._info

        rule.pop("links", None)
        return self.dict2columns(dict(sorted(rule.items())))


class SetBaremetalInspectionRule(command.Command):
    """Set baremetal rule properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".SetBaremetalInspectionRule")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'rule',
            metavar='<rule>',
            help=_("UUID of the inspection rule")
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('Set a brief explanation about the rule.')
        )
        parser.add_argument(
            '--priority',
            metavar='<priority>',
            help=_("Specifies the rule's precedence level during execution.")
        )
        parser.add_argument(
            '--phase',
            metavar='<phase>',
            help=_('Specifies the processing phase when the rule should run.')
        )
        parser.add_argument(
            '--conditions',
            metavar="<conditions>",
            help=_('Conditions under which the rule should be triggered.')
        )
        parser.add_argument(
            '--actions',
            metavar="<actions>",
            help=_('Actions to be executed when the rule conditions are met.')
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        if parsed_args.description:
            description = ["description=%s" % parsed_args.description]
            properties.extend(utils.args_array_to_patch('add', description))
        if parsed_args.priority:
            priority = ["priority=%s" % parsed_args.priority]
            properties.extend(utils.args_array_to_patch('add', priority))
        if parsed_args.phase:
            phase = ["phase=%s" % parsed_args.phase]
            properties.extend(utils.args_array_to_patch('add', phase))
        if parsed_args.actions:
            actions = utils.handle_json_arg(parsed_args.actions,
                                            'rule actions')
            actions = ["actions=%s" % json.dumps(actions)]
            properties.extend(utils.args_array_to_patch('add', actions))
        if parsed_args.conditions:
            conditions = utils.handle_json_arg(parsed_args.conditions,
                                               'rule conditions')
            conditions = ["conditions=%s" % json.dumps(conditions)]
            properties.extend(utils.args_array_to_patch('add', conditions))

        if properties:
            baremetal_client.inspection_rule.update(parsed_args.rule,
                                                    properties)
        else:
            self.log.warning("Please specify what to set.")


class UnsetBaremetalInspectionRule(command.Command):
    """Unset baremetal rule properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".UnsetBaremetalInspectionRule")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'rule',
            metavar='<rule>',
            help=_("UUID of the inspection rule")
        )
        parser.add_argument(
            '--description',
            dest='description',
            action='store_true',
            help=_('Unset a brief explanation about the rule.')
        )
        parser.add_argument(
            '--priority',
            dest='priority',
            action='store_true',
            help=_("Specifies the rule's precedence level during execution.")
        )
        parser.add_argument(
            '--phase',
            dest='phase',
            action='store_true',
            help=_('Specifies the processing phase when the rule should run.')
        )
        parser.add_argument(
            '--condition',
            metavar="<key>",
            action='append',
            help=_('Condition to unset on this baremetal inspection rule '
                   '(repeat option to unset multiple conditions).')
        )
        parser.add_argument(
            '--action',
            metavar="<key>",
            action='append',
            help=_('Action to unset on this baremetal inspection rule '
                   '(repeat option to unset multiple actions).')
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        field_list = ['description', 'priority', 'phase']
        for field in field_list:
            if getattr(parsed_args, field):
                properties.extend(utils.args_array_to_patch('remove', [field]))

        if parsed_args.action:
            properties.extend(utils.args_array_to_patch('remove',
                              ['action/' + x for x in parsed_args.action]))
        if parsed_args.condition:
            properties.extend(
                utils.args_array_to_patch(
                    'remove',
                    ['condition/' + x for x in parsed_args.condition]))

        if properties:
            baremetal_client.inspection_rule.update(parsed_args.rule,
                                                    properties)
        else:
            self.log.warning("Please specify what to unset.")


class DeleteBaremetalInspectionRule(command.Command):
    """Delete rule(s)."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".DeleteBaremetalInspectionRule")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "rules",
            metavar="<rule>",
            nargs="+",
            help=_("UUID(s) of the rule(s) to delete."),
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        failures: list[str] = []
        if parsed_args.rules == 'all':
            try:
                baremetal_client.inspection_rule.delete()
                print(_('Deleted all rules.'))
            except exc.ClientException as e:
                raise exc.ClientException(
                    _("Failed to delete all rules: %s") % e)
        else:
            for rule in parsed_args.rules:
                try:
                    baremetal_client.inspection_rule.delete(rule)
                    print(_('Deleted rule %s') % rule)
                except exc.ClientException as e:
                    failures.append(_("Failed to delete rule "
                                      "%(rule)s: %(error)s")
                                    % {'rule': rule, 'error': e})

        if failures:
            raise exc.ClientException("\n".join(failures))


class ListBaremetalInspectionRule(command.Lister):
    """List baremetal rules."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalInspectionRule")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of rules to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Baremetal API Service.')
        )
        parser.add_argument(
            '--marker',
            metavar='<rule>',
            help=_('InspectionRule UUID (for example, of the last rule '
                   'in the list from a previous request). Returns '
                   'the list of rules after this UUID.')
        )
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            help=_('Sort output by specified rule fields and '
                   'directions (asc or desc) (default: asc). Multiple fields '
                   'and directions can be specified, separated by comma.')
        )
        display_group = parser.add_mutually_exclusive_group()
        display_group.add_argument(
            '--long',
            dest='detail',
            action='store_true',
            default=False,
            help=_("Show detailed information about rules.")
        )
        display_group.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.INSPECTION_RULE_DETAILED_RESOURCE.fields,
            help=_("One or more rule fields. Only these fields "
                   "will be fetched from the server. Can not be used when "
                   "'--long' is specified.")
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        client = manager.baremetal

        columns = res_fields.INSPECTION_RULE_RESOURCE.fields

        params: dict[str, object] = {}
        if parsed_args.limit is not None and parsed_args.limit < 0:
            raise exc.CommandError(
                _('Expected non-negative --limit, got %s') %
                parsed_args.limit)
        params['limit'] = parsed_args.limit
        params['marker'] = parsed_args.marker

        if parsed_args.detail:
            params['detail'] = parsed_args.detail
            columns = res_fields.INSPECTION_RULE_DETAILED_RESOURCE.fields

        elif parsed_args.fields:
            params['detail'] = False
            fields = itertools.chain.from_iterable(parsed_args.fields)
            resource = res_fields.Resource(list(fields))
            columns = resource.fields
            params['fields'] = columns

        self.log.debug("params(%s)", params)
        data = client.inspection_rule.list(**params)

        data = oscutils.sort_items(data, parsed_args.sort)

        return (columns,
                (oscutils.get_item_properties(s, columns) for s in data))



#
#   Copyright 2016 Mirantis, Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import copy
from unittest import mock

from osc_lib.tests import utils as osctestutils

from ironicclient.osc.v1 import baremetal_portgroup
from ironicclient.tests.unit.osc.v1 import fakes as baremetal_fakes


class TestBaremetalPortGroup(baremetal_fakes.TestBaremetal):

    def setUp(self) -> None:
        super(TestBaremetalPortGroup, self).setUp()

        self.baremetal_mock = self.app.client_manager.baremetal
        self.baremetal_mock.reset_mock()


class TestCreateBaremetalPortGroup(TestBaremetalPortGroup):

    def setUp(self) -> None:
        super(TestCreateBaremetalPortGroup, self).setUp()

        self.baremetal_mock.portgroup.create.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.PORTGROUP),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = baremetal_portgroup.CreateBaremetalPortGroup(self.app, None)

    def test_baremetal_portgroup_create(self) -> None:
        arglist = [
            '--node', baremetal_fakes.baremetal_uuid,
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'node_uuid': baremetal_fakes.baremetal_uuid,
        }

        self.baremetal_mock.portgroup.create.assert_called_once_with(**args)

    def test_baremetal_portgroup_create_name_address_uuid(self) -> None:
        arglist = [
            '--address', baremetal_fakes.baremetal_portgroup_address,
            '--node', baremetal_fakes.baremetal_uuid,
            '--name', baremetal_fakes.baremetal_portgroup_name,
            '--uuid', baremetal_fakes.baremetal_portgroup_uuid,
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('address', baremetal_fakes.baremetal_portgroup_address),
            ('name', baremetal_fakes.baremetal_portgroup_name),
            ('uuid', baremetal_fakes.baremetal_portgroup_uuid),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'address': baremetal_fakes.baremetal_portgroup_address,
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'name': baremetal_fakes.baremetal_portgroup_name,
            'uuid': baremetal_fakes.baremetal_portgroup_uuid,
        }

        self.baremetal_mock.portgroup.create.assert_called_once_with(**args)

    def test_baremetal_portgroup_create_support_standalone_ports(self) -> None:
        arglist = [
            '--address', baremetal_fakes.baremetal_portgroup_address,
            '--node', baremetal_fakes.baremetal_uuid,
            '--support-standalone-ports'
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('address', baremetal_fakes.baremetal_portgroup_address),
            ('support_standalone_ports', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'address': baremetal_fakes.baremetal_portgroup_address,
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'standalone_ports_supported': True,
        }

        self.baremetal_mock.portgroup.create.assert_called_once_with(**args)

    def test_baremetal_portgroup_create_unsupport_standalone_ports(
            self) -> None:
        arglist = [
            '--address', baremetal_fakes.baremetal_portgroup_address,
            '--node', baremetal_fakes.baremetal_uuid,
            '--unsupport-standalone-ports'
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('address', baremetal_fakes.baremetal_portgroup_address),
            ('unsupport_standalone_ports', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'address': baremetal_fakes.baremetal_portgroup_address,
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'standalone_ports_supported': False,
        }

        self.baremetal_mock.portgroup.create.assert_called_once_with(**args)

    def test_baremetal_portgroup_create_name_extras(self) -> None:
        arglist = [
            '--address', baremetal_fakes.baremetal_portgroup_address,
            '--node', baremetal_fakes.baremetal_uuid,
            '--name', baremetal_fakes.baremetal_portgroup_name,
            '--extra', 'key1=value1',
            '--extra', 'key2=value2'
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('address', baremetal_fakes.baremetal_portgroup_address),
            ('name', baremetal_fakes.baremetal_portgroup_name),
            ('extra', ['key1=value1', 'key2=value2'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'address': baremetal_fakes.baremetal_portgroup_address,
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'name': baremetal_fakes.baremetal_portgroup_name,
            'extra': baremetal_fakes.baremetal_portgroup_extra
        }

        self.baremetal_mock.portgroup.create.assert_called_once_with(**args)

    def test_baremetal_portgroup_create_mode_properties(self) -> None:
        arglist = [
            '--node', baremetal_fakes.baremetal_uuid,
            '--mode', baremetal_fakes.baremetal_portgroup_mode,
            '--property', 'key1=value11',
            '--property', 'key2=value22'
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('mode', baremetal_fakes.baremetal_portgroup_mode),
            ('properties', ['key1=value11', 'key2=value22'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'mode': baremetal_fakes.baremetal_portgroup_mode,
            'properties': baremetal_fakes.baremetal_portgroup_properties
        }

        self.baremetal_mock.portgroup.create.assert_called_once_with(**args)

    def test_baremetal_portgroup_create_no_options(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_portgroup_create_physical_network(self) -> None:
        arglist = [
            '--address', baremetal_fakes.baremetal_portgroup_address,
            '--node', baremetal_fakes.baremetal_uuid,
            '--physical-network',
            baremetal_fakes.baremetal_port_physical_network,
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('address', baremetal_fakes.baremetal_portgroup_address),
            ('physical_network',
             baremetal_fakes.baremetal_port_physical_network),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'address': baremetal_fakes.baremetal_portgroup_address,
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'physical_network':
            baremetal_fakes.baremetal_port_physical_network,
        }

        self.baremetal_mock.portgroup.create.assert_called_once_with(**args)

    def test_baremetal_portgroup_create_category(self) -> None:
        arglist = [
            '--address', baremetal_fakes.baremetal_portgroup_address,
            '--node', baremetal_fakes.baremetal_uuid,
            '--category', baremetal_fakes.baremetal_portgroup_category,
        ]

        verifylist = [
            ('node_uuid', baremetal_fakes.baremetal_uuid),
            ('address', baremetal_fakes.baremetal_portgroup_address),
            ('category', baremetal_fakes.baremetal_portgroup_category),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        self.cmd.take_action(parsed_args)

        # Set expected values
        args = {
            'address': baremetal_fakes.baremetal_portgroup_address,
            'node_uuid': baremetal_fakes.baremetal_uuid,
            'category': baremetal_fakes.baremetal_portgroup_category,
        }

        self.baremetal_mock.portgroup.create.assert_called_once_with(**args)


class TestShowBaremetalPortGroup(TestBaremetalPortGroup):

    def setUp(self) -> None:
        super(TestShowBaremetalPortGroup, self).setUp()

        self.baremetal_mock.portgroup.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.PORTGROUP),
                loaded=True))

        self.baremetal_mock.portgroup.get_by_address.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.PORTGROUP),
                loaded=True))

        self.cmd = baremetal_portgroup.ShowBaremetalPortGroup(self.app, None)

    def test_baremetal_portgroup_show(self) -> None:
        arglist = ['ppp-gggggg-pppp']
        verifylist = [('portgroup', baremetal_fakes.baremetal_portgroup_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        args = ['ppp-gggggg-pppp']
        self.baremetal_mock.portgroup.get.assert_called_with(*args,
                                                             fields=None)

        collist = ('address', 'category', 'extra', 'mode', 'name',
                   'node_uuid', 'physical_network', 'properties',
                   'uuid')
        self.assertEqual(collist, columns)

        datalist = (
            baremetal_fakes.baremetal_portgroup_address,
            baremetal_fakes.baremetal_portgroup_category,
            baremetal_fakes.baremetal_portgroup_extra,
            baremetal_fakes.baremetal_portgroup_mode,
            baremetal_fakes.baremetal_portgroup_name,
            baremetal_fakes.baremetal_uuid,
            baremetal_fakes.baremetal_port_physical_network,
            baremetal_fakes.baremetal_portgroup_properties,
            baremetal_fakes.baremetal_portgroup_uuid,
        )

        self.assertEqual(datalist, tuple(data))

    def test_baremetal_portgroup_show_address(self) -> None:
        arglist = ['--address', baremetal_fakes.baremetal_portgroup_address]
        verifylist = [('address', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        args = {baremetal_fakes.baremetal_portgroup_address}
        self.baremetal_mock.portgroup.get_by_address.assert_called_with(
            *args, fields=None)

    def test_baremetal_portgroup_show_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalPortGroupList(TestBaremetalPortGroup):
    def setUp(self) -> None:
        super(TestBaremetalPortGroupList, self).setUp()

        self.baremetal_mock.portgroup.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.PORTGROUP),
                loaded=True)
        ]
        self.cmd = baremetal_portgroup.ListBaremetalPortGroup(self.app, None)

    def test_baremetal_portgroup_list(self) -> None:
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None}
        self.baremetal_mock.portgroup.list.assert_called_with(**kwargs)

        collist = (
            "uuid",
            "address",
            "name")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_portgroup_uuid,
                     baremetal_fakes.baremetal_portgroup_address,
                     baremetal_fakes.baremetal_portgroup_name),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_portgroup_list_address(self) -> None:
        arglist = ['--address', baremetal_fakes.baremetal_portgroup_address]
        verifylist = [('address', baremetal_fakes.baremetal_portgroup_address)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'address': baremetal_fakes.baremetal_portgroup_address,
            'marker': None,
            'limit': None}
        self.baremetal_mock.portgroup.list.assert_called_with(**kwargs)

        collist = (
            "uuid",
            "address",
            "name")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_portgroup_uuid,
                     baremetal_fakes.baremetal_portgroup_address,
                     baremetal_fakes.baremetal_portgroup_name),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_portgroup_list_node(self) -> None:
        arglist = ['--node', baremetal_fakes.baremetal_uuid]
        verifylist = [('node', baremetal_fakes.baremetal_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'node': baremetal_fakes.baremetal_uuid,
            'marker': None,
            'limit': None}
        self.baremetal_mock.portgroup.list.assert_called_with(**kwargs)

        collist = (
            "uuid",
            "address",
            "name")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_portgroup_uuid,
                     baremetal_fakes.baremetal_portgroup_address,
                     baremetal_fakes.baremetal_portgroup_name),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_portgroup_list_long(self) -> None:
        arglist = ['--long']
        verifylist = [('detail', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'detail': True,
            'marker': None,
            'limit': None,
        }
        self.baremetal_mock.portgroup.list.assert_called_with(**kwargs)

        collist = ('uuid', 'address', 'created_at', 'extra',
                   'standalone_ports_supported', 'node_uuid', 'name',
                   'updated_at', 'internal_info', 'mode', 'properties',
                   'physical_network', 'category')

        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_portgroup_uuid,
                     baremetal_fakes.baremetal_portgroup_address,
                     '',
                     baremetal_fakes.baremetal_portgroup_extra,
                     '',
                     baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_portgroup_name,
                     '',
                     '',
                     baremetal_fakes.baremetal_portgroup_mode,
                     baremetal_fakes.baremetal_portgroup_properties,
                     baremetal_fakes.baremetal_port_physical_network,
                     baremetal_fakes.baremetal_portgroup_category,
                     ),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_portgroup_list_fields(self) -> None:
        arglist = ['--fields', 'uuid', 'address']
        verifylist = [('fields', [['uuid', 'address']])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None,
            'detail': False,
            'fields': ('uuid', 'address')
        }
        self.baremetal_mock.portgroup.list.assert_called_with(**kwargs)

    def test_baremetal_portgroup_list_fields_multiple(self) -> None:
        arglist = ['--fields', 'uuid', 'address', '--fields', 'extra']
        verifylist = [('fields', [['uuid', 'address'], ['extra']])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None,
            'detail': False,
            'fields': ('uuid', 'address', 'extra')
        }
        self.baremetal_mock.portgroup.list.assert_called_with(**kwargs)

    def test_baremetal_portgroup_list_invalid_fields(self) -> None:
        arglist = ['--fields', 'uuid', 'invalid']
        verifylist = [('fields', [['uuid', 'invalid']])]
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_portgroup_list_by_shards(self) -> None:
        arglist = ['--shards', 'shard1', 'shard2']
        verifylist = [('shards', ['shard1', 'shard2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'shards': ['shard1', 'shard2'],
            'marker': None,
            'limit': None}
        self.baremetal_mock.portgroup.list.assert_called_with(**kwargs)

        collist = (
            "uuid",
            "address",
            "name")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_portgroup_uuid,
                     baremetal_fakes.baremetal_portgroup_address,
                     baremetal_fakes.baremetal_portgroup_name),)
        self.assertEqual(datalist, tuple(data))


class TestBaremetalPortGroupDelete(TestBaremetalPortGroup):

    def setUp(self) -> None:
        super(TestBaremetalPortGroupDelete, self).setUp()

        self.baremetal_mock.portgroup.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.PORTGROUP),
                loaded=True))

        self.cmd = baremetal_portgroup.DeleteBaremetalPortGroup(self.app, None)

    def test_baremetal_portgroup_delete(self) -> None:
        arglist = [baremetal_fakes.baremetal_portgroup_uuid]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        args = baremetal_fakes.baremetal_portgroup_uuid
        self.baremetal_mock.portgroup.delete.assert_called_with(args)

    def test_baremetal_portgroup_delete_multiple(self) -> None:
        arglist = [baremetal_fakes.baremetal_portgroup_uuid,
                   baremetal_fakes.baremetal_portgroup_name]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        args = [baremetal_fakes.baremetal_portgroup_uuid,
                baremetal_fakes.baremetal_portgroup_name]
        self.baremetal_mock.portgroup.delete.assert_has_calls(
            [mock.call(x) for x in args])
        self.assertEqual(2, self.baremetal_mock.portgroup.delete.call_count)

    def test_baremetal_portgroup_delete_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalPortGroupSet(TestBaremetalPortGroup):
    def setUp(self) -> None:
        super(TestBaremetalPortGroupSet, self).setUp()

        self.baremetal_mock.portgroup.update.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.PORTGROUP),
                loaded=True))

        self.cmd = baremetal_portgroup.SetBaremetalPortGroup(self.app, None)

    def test_baremetal_portgroup_set_name(self) -> None:
        new_portgroup_name = 'New-Portgroup-name'
        arglist = [
            baremetal_fakes.baremetal_portgroup_uuid,
            '--name', new_portgroup_name]
        verifylist = [
            ('portgroup', baremetal_fakes.baremetal_portgroup_uuid),
            ('name', new_portgroup_name)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            baremetal_fakes.baremetal_portgroup_uuid,
            [{'path': '/name', 'value': new_portgroup_name, 'op': 'add'}])

    def test_baremetal_portgroup_set_address(self) -> None:
        new_portgroup_address = '00:22:44:66:88:00'
        arglist = [
            baremetal_fakes.baremetal_portgroup_uuid,
            '--address', new_portgroup_address]
        verifylist = [
            ('portgroup', baremetal_fakes.baremetal_portgroup_uuid),
            ('address', new_portgroup_address)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            baremetal_fakes.baremetal_portgroup_uuid,
            [{'path': '/address', 'value': new_portgroup_address,
              'op': 'add'}])

    def test_baremetal_portgroup_set_mode(self) -> None:
        new_portgroup_mode = '802.3ad'
        arglist = [
            baremetal_fakes.baremetal_portgroup_uuid,
            '--mode', new_portgroup_mode]
        verifylist = [
            ('portgroup', baremetal_fakes.baremetal_portgroup_uuid),
            ('mode', new_portgroup_mode)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            baremetal_fakes.baremetal_portgroup_uuid,
            [{'path': '/mode', 'value': new_portgroup_mode,
              'op': 'add'}])

    def test_baremetal_portgroup_set_mode_int(self) -> None:
        new_portgroup_mode = '4'
        arglist = [
            baremetal_fakes.baremetal_portgroup_uuid,
            '--mode', new_portgroup_mode]
        verifylist = [
            ('portgroup', baremetal_fakes.baremetal_portgroup_uuid),
            ('mode', new_portgroup_mode)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            baremetal_fakes.baremetal_portgroup_uuid,
            [{'path': '/mode', 'value': new_portgroup_mode,
              'op': 'add'}])

    def test_baremetal_portgroup_set_node_uuid(self) -> None:
        new_node_uuid = 'nnnnnn-uuuuuuuu'
        arglist = [
            baremetal_fakes.baremetal_portgroup_uuid,
            '--node', new_node_uuid]
        verifylist = [
            ('portgroup', baremetal_fakes.baremetal_portgroup_uuid),
            ('node_uuid', new_node_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            baremetal_fakes.baremetal_portgroup_uuid,
            [{'path': '/node_uuid', 'value': new_node_uuid,
              'op': 'add'}])

    def test_baremetal_portgroup_set_support_standalone_ports(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_portgroup_uuid,
            '--support-standalone-ports']
        verifylist = [
            ('portgroup', baremetal_fakes.baremetal_portgroup_uuid),
            ('support_standalone_ports', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            baremetal_fakes.baremetal_portgroup_uuid,
            [{'path': '/standalone_ports_supported', 'value': 'True',
              'op': 'add'}])

    def test_baremetal_portgroup_set_unsupport_standalone_ports(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_portgroup_uuid,
            '--unsupport-standalone-ports']
        verifylist = [
            ('portgroup', baremetal_fakes.baremetal_portgroup_uuid),
            ('unsupport_standalone_ports', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            baremetal_fakes.baremetal_portgroup_uuid,
            [{'path': '/standalone_ports_supported', 'value': 'False',
              'op': 'add'}])

    def test_baremetal_set_extra(self) -> None:
        arglist = ['portgroup', '--extra', 'foo=bar']
        verifylist = [('portgroup', 'portgroup'),
                      ('extra', ['foo=bar'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/extra/foo', 'value': 'bar', 'op': 'add'}])

    def test_baremetal_portgroup_set_multiple_extras(self) -> None:
        arglist = ['portgroup',
                   '--extra', 'key1=val1',
                   '--extra', 'key2=val2']
        verifylist = [('portgroup', 'portgroup'),
                      ('extra', ['key1=val1', 'key2=val2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/extra/key1', 'value': 'val1', 'op': 'add'},
             {'path': '/extra/key2', 'value': 'val2', 'op': 'add'}])

    def test_baremetal_portgroup_set_multiple_properties(self) -> None:
        arglist = ['portgroup',
                   '--property', 'key3=val3',
                   '--property', 'key4=val4']
        verifylist = [('portgroup', 'portgroup'),
                      ('properties', ['key3=val3', 'key4=val4'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/properties/key3', 'value': 'val3', 'op': 'add'},
             {'path': '/properties/key4', 'value': 'val4', 'op': 'add'}])

    def test_baremetal_portgroup_set_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_portgroup_set_no_property(self) -> None:
        uuid = baremetal_fakes.baremetal_portgroup_uuid
        arglist = [uuid]
        verifylist = [('portgroup', uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.assertFalse(self.baremetal_mock.portgroup.update.called)

    def test_baremetal_portgroup_set_physical_network(self) -> None:
        new_portgroup_physnet = 'physnet2'
        arglist = [
            baremetal_fakes.baremetal_portgroup_uuid,
            '--physical-network', new_portgroup_physnet]
        verifylist = [
            ('portgroup', baremetal_fakes.baremetal_portgroup_uuid),
            ('physical_network', new_portgroup_physnet)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            baremetal_fakes.baremetal_portgroup_uuid,
            [{'path': '/physical_network', 'value': new_portgroup_physnet,
              'op': 'add'}])

    def test_baremetal_portgroup_set_category(self) -> None:
        new_portgroup_category = 'Purple'
        arglist = [
            baremetal_fakes.baremetal_portgroup_uuid,
            '--category', new_portgroup_category]
        verifylist = [
            ('portgroup', baremetal_fakes.baremetal_portgroup_uuid),
            ('category', new_portgroup_category)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            baremetal_fakes.baremetal_portgroup_uuid,
            [{'path': '/category', 'value': new_portgroup_category,
              'op': 'add'}])


class TestBaremetalPortGroupUnset(TestBaremetalPortGroup):
    def setUp(self) -> None:
        super(TestBaremetalPortGroupUnset, self).setUp()

        self.baremetal_mock.portgroup.update.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.PORTGROUP),
                loaded=True))

        self.cmd = baremetal_portgroup.UnsetBaremetalPortGroup(self.app, None)

    def test_baremetal_portgroup_unset_extra(self) -> None:
        arglist = ['portgroup', '--extra', 'key1']
        verifylist = [('portgroup', 'portgroup'),
                      ('extra', ['key1'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/extra/key1', 'op': 'remove'}])

    def test_baremetal_portgroup_unset_name(self) -> None:
        arglist = ['portgroup', '--name']
        verifylist = [('name', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/name', 'op': 'remove'}])

    def test_baremetal_portgroup_unset_address(self) -> None:
        arglist = ['portgroup', '--address']
        verifylist = [('address', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/address', 'op': 'remove'}])

    def test_baremetal_portgroup_unset_multiple_extras(self) -> None:
        arglist = ['portgroup',
                   '--extra', 'key1',
                   '--extra', 'key2']
        verifylist = [('portgroup', 'portgroup'),
                      ('extra', ['key1', 'key2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/extra/key1', 'op': 'remove'},
             {'path': '/extra/key2', 'op': 'remove'}])

    def test_baremetal_portgroup_unset_multiple_properties(self) -> None:
        arglist = ['portgroup',
                   '--property', 'key1',
                   '--property', 'key2']
        verifylist = [('portgroup', 'portgroup'),
                      ('properties', ['key1', 'key2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/properties/key1', 'op': 'remove'},
             {'path': '/properties/key2', 'op': 'remove'}])

    def test_baremetal_portgroup_unset_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_portgroup_unset_no_property(self) -> None:
        uuid = baremetal_fakes.baremetal_portgroup_uuid
        arglist = [uuid]
        verifylist = [('portgroup', uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.assertFalse(self.baremetal_mock.portgroup.update.called)

    def test_baremetal_portgroup_unset_physical_network(self) -> None:
        arglist = ['portgroup', '--physical-network']
        verifylist = [('physical_network', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/physical_network', 'op': 'remove'}])

    def test_baremetal_portgroup_unset_category(self) -> None:
        arglist = ['portgroup', '--category']
        verifylist = [('category', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.portgroup.update.assert_called_once_with(
            'portgroup',
            [{'path': '/category', 'op': 'remove'}])



#
#   Copyright 2016 Intel Corporation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
import itertools
import logging
from typing import Any

from osc_lib import utils as oscutils

from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc
from ironicclient.osc import command
from ironicclient.v1 import resource_fields as res_fields


class CreateBaremetalChassis(command.ShowOne):
    """Create a new chassis."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".CreateBaremetalChassis")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            '--description',
            dest='description',
            metavar='<description>',
            help=_('Description for the chassis')
        )
        parser.add_argument(
            '--extra',
            metavar='<key=value>',
            action='append',
            help=_("Record arbitrary key/value metadata. "
                   "Can be specified multiple times.")
        )
        parser.add_argument(
            '--uuid',
            metavar='<uuid>',
            help=_("Unique UUID of the chassis")
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        field_list: list[str] = ['description', 'extra', 'uuid']
        fields = dict((k, v) for (k, v) in vars(parsed_args).items()
                      if k in field_list and v is not None)
        fields = utils.args_array_to_dict(fields, 'extra')
        chassis = baremetal_client.chassis.create(**fields)._info

        chassis.pop('links', None)
        chassis.pop('nodes', None)

        return self.dict2columns(chassis)


class DeleteBaremetalChassis(command.Command):
    """Delete a chassis."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".DeleteBaremetalChassis")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "chassis",
            metavar="<chassis>",
            nargs="+",
            help=_("UUIDs of chassis to delete")
        )

        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        failures: list[str] = []
        for chassis in parsed_args.chassis:
            try:
                baremetal_client.chassis.delete(chassis)
                print(_('Deleted chassis %s') % chassis)
            except exc.ClientException as e:
                failures.append(_("Failed to delete chassis %(chassis)s: "
                                  "%(error)s")
                                % {'chassis': chassis, 'error': e})

        if failures:
            raise exc.ClientException("\n".join(failures))


class ListBaremetalChassis(command.Lister):
    """List the chassis."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalChassis")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        display_group = parser.add_mutually_exclusive_group(required=False)
        display_group.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.CHASSIS_DETAILED_RESOURCE.fields,
            help=_("One or more chassis fields. Only these fields will be "
                   "fetched from the server. Cannot be used when '--long' is "
                   "specified.")
        )
        parser.add_argument(
            '--limit',
            metavar='<limit>',
            type=int,
            help=_('Maximum number of chassis to return per request, '
                   '0 for no limit. Default is the maximum number used '
                   'by the Baremetal API Service.')
        )
        display_group.add_argument(
            '--long',
            default=False,
            action='store_true',
            help=_("Show detailed information about the chassis")
        )
        parser.add_argument(
            '--marker',
            metavar='<chassis>',
            help=_('Chassis UUID (for example, of the last chassis in the '
                   'list from a previous request). Returns the list of '
                   'chassis after this UUID.')
        )
        parser.add_argument(
            '--sort',
            metavar="<key>[:<direction>]",
            help=_('Sort output by specified chassis fields and directions '
                   '(asc or desc) (default: asc). Multiple fields and '
                   'directions can be specified, separated by comma.')
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        client = manager.baremetal

        columns = res_fields.CHASSIS_RESOURCE.fields

        params: dict[str, object] = {}
        if parsed_args.limit is not None and parsed_args.limit < 0:
            raise exc.CommandError(
                _('Expected non-negative --limit, got %s') %
                parsed_args.limit)
        params['limit'] = parsed_args.limit
        params['marker'] = parsed_args.marker
        if parsed_args.long:
            params['detail'] = parsed_args.long
            columns = res_fields.CHASSIS_DETAILED_RESOURCE.fields
        elif parsed_args.fields:
            params['detail'] = False
            fields = itertools.chain.from_iterable(parsed_args.fields)
            resource = res_fields.Resource(list(fields))
            columns = resource.fields
            params['fields'] = columns

        self.log.debug("params(%s)", params)
        data = client.chassis.list(**params)

        data = oscutils.sort_items(data, parsed_args.sort)

        return (columns,
                (oscutils.get_item_properties(s, columns, formatters={
                    'Properties': utils.HashColumn},) for s in data))


class SetBaremetalChassis(command.Command):
    """Set chassis properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".SetBaremetalChassis")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'chassis',
            metavar='<chassis>',
            help=_("UUID of the chassis")
        )
        parser.add_argument(
            "--description",
            metavar="<description>",
            help=_("Set the description of the chassis")
        )
        parser.add_argument(
            "--extra",
            metavar="<key=value>",
            action='append',
            help=_('Extra to set on this chassis '
                   '(repeat option to set multiple extras)')
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        if parsed_args.description:
            description = ["description=%s" % parsed_args.description]
            properties.extend(utils.args_array_to_patch(
                'add', description))
        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch(
                'add', ['extra/' + x for x in parsed_args.extra]))

        if properties:
            baremetal_client.chassis.update(parsed_args.chassis, properties)
        else:
            self.log.warning("Please specify what to set.")


class ShowBaremetalChassis(command.ShowOne):
    """Show chassis details."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ShowBaremetalChassis")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            "chassis",
            metavar="<chassis>",
            help=_("UUID of the chassis")
        )
        parser.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            choices=res_fields.CHASSIS_DETAILED_RESOURCE.fields,
            default=[],
            help=_("One or more chassis fields. Only these fields will be "
                   "fetched from the server.")
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[tuple[str, ...], tuple[Any, ...]]:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal
        fields: list[str] | None
        fields = (
            list(itertools.chain.from_iterable(parsed_args.fields))
            or None)
        chassis = baremetal_client.chassis.get(parsed_args.chassis,
                                               fields=fields)._info
        chassis.pop("links", None)
        chassis.pop("nodes", None)

        return self.dict2columns(dict(sorted(chassis.items())))


class UnsetBaremetalChassis(command.Command):
    """Unset chassis properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".UnsetBaremetalChassis")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)

        parser.add_argument(
            'chassis',
            metavar='<chassis>',
            help=_("UUID of the chassis")
        )
        parser.add_argument(
            '--description',
            action='store_true',
            default=False,
            help=_('Clear the chassis description')
        )
        parser.add_argument(
            "--extra",
            metavar="<key>",
            action='append',
            help=_('Extra to unset on this chassis '
                   '(repeat option to unset multiple extras)')
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        self.log.debug("take_action(%s)", parsed_args)

        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        properties: list[dict[str, Any]] = []
        if parsed_args.description:
            properties.extend(utils.args_array_to_patch('remove',
                              ['description']))
        if parsed_args.extra:
            properties.extend(utils.args_array_to_patch('remove',
                              ['extra/' + x for x in parsed_args.extra]))

        if properties:
            baremetal_client.chassis.update(parsed_args.chassis, properties)
        else:
            self.log.warning("Please specify what to unset.")



# Copyright 2010 Jacob Kaplan-Moss
# Copyright 2011 OpenStack Foundation
# Copyright 2012 Grid Dynamics
# Copyright 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base utilities to build API operation managers and objects on top of.
"""

from __future__ import annotations

# E1102: %s is not callable
# pylint: disable=E1102

import copy
from typing import Any, Callable, Protocol

from oslo_utils import strutils


class ManagerProtocol(Protocol):
    """Minimal protocol for Resource's manager dependency."""

    client: Any


class Resource(object):
    """Base class for OpenStack resources (tenant, user, etc.).

    This is pretty much just a bag for attributes.
    """

    HUMAN_ID: bool = False
    NAME_ATTR: str = "name"

    def __init__(
        self,
        manager: ManagerProtocol,
        info: dict[str, Any],
        loaded: bool = False,
    ) -> None:
        """Populate and bind to a manager.

        :param manager: BaseManager object
        :param info: dictionary representing resource attributes
        :param loaded: prevent lazy-loading if set to True
        """
        self.manager = manager
        self._info = info
        self._add_details(info)
        self._loaded = loaded

    def __repr__(self) -> str:
        reprkeys = sorted(
            k for k in self.__dict__.keys() if k[0] != "_" and k != "manager"
        )
        info = ", ".join("%s=%s" % (k, getattr(self, k)) for k in reprkeys)
        return "<%s %s>" % (self.__class__.__name__, info)

    @property
    def human_id(self) -> str | None:
        """Human-readable ID which can be used for bash completion."""
        if self.HUMAN_ID:
            name = getattr(self, self.NAME_ATTR, None)
            if name is not None:
                return str(strutils.to_slug(name))
        return None

    def _add_details(self, info: dict[str, Any]) -> None:
        for k, v in info.items():
            try:
                setattr(self, k, v)
                self._info[k] = v
            except AttributeError:
                # In this case we already defined the attribute on the class
                pass

    def __getattr__(self, k: str) -> Any:
        if k not in self.__dict__:
            # NOTE(bcwaldon): disallow lazy-loading if already loaded once
            if not self.is_loaded():
                self.get()
                return self.__getattr__(k)

            raise AttributeError(k)
        else:
            return self.__dict__[k]

    def get(self) -> None:
        """Support for lazy loading details.

        Some clients, such as novaclient have the option to lazy load the
        details, details which can be loaded with this function.
        """
        # set_loaded() first ... so if we have to bail, we know we tried.
        self.set_loaded(True)
        get_fn: Callable[..., Resource] | None = getattr(
            self.manager, "get", None,
        )
        if get_fn is None:
            return

        new = get_fn(self.id)
        if new:
            self._add_details(new._info)
            self._add_details(
                {"x_request_id": self.manager.client.last_request_id}
            )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Resource):
            return NotImplemented
        # two resources of different types are not equal
        if not isinstance(other, self.__class__):
            return False
        return self._info == other._info

    def is_loaded(self) -> bool:
        return self._loaded

    def set_loaded(self, val: bool) -> None:
        self._loaded = val

    def to_dict(self) -> dict[str, Any]:
        return copy.deepcopy(self._info)



#  Copyright (c) 2016 Mirantis, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
#
from __future__ import annotations

import argparse
from collections.abc import Iterable, Sequence
import itertools
import logging
from typing import Any, cast

from osc_lib import utils as oscutils

from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient.osc import command
from ironicclient.v1 import resource_fields as res_fields
from ironicclient.v1 import utils as v1_utils


class ListBaremetalDriver(command.Lister):
    """List the enabled drivers."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalDriver")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--type',
            metavar='<type>',
            choices=["classic", "dynamic"],
            help='Type of driver ("classic" or "dynamic"). '
                 'The default is to list all of them.'
        )
        display_group = parser.add_mutually_exclusive_group()
        display_group.add_argument(
            '--long',
            action='store_true',
            default=None,
            help="Show detailed information about the drivers.")
        display_group.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.DRIVER_DETAILED_RESOURCE.fields,
            help=_("One or more node fields. Only these fields will be "
                   "fetched from the server. Can not be used when '--long' "
                   "is specified."))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        client = manager.baremetal

        params: dict[str, object] = {
            'driver_type': parsed_args.type,
            'detail': parsed_args.long,
        }
        if parsed_args.long:
            columns = res_fields.DRIVER_DETAILED_RESOURCE.fields
        elif parsed_args.fields:
            fields = itertools.chain.from_iterable(parsed_args.fields)
            resource = res_fields.Resource(list(fields))
            columns = resource.fields
            params['fields'] = columns
        else:
            columns = res_fields.DRIVER_RESOURCE.fields

        drivers = client.driver.list(**params)
        drivers = oscutils.sort_items(drivers, 'name')

        # For list-type properties, show the values as comma separated
        # strings. It's easier to read.
        data = [utils.convert_list_props_to_comma_separated(d._info)
                for d in drivers]

        return (columns,
                (oscutils.get_dict_properties(s, columns) for s in data))


class ListBaremetalDriverProperty(command.Lister):
    """List the driver properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalDriverProperty")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'driver',
            metavar='<driver>',
            help='Name of the driver.')
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        driver_properties = baremetal_client.driver.properties(
            parsed_args.driver)
        columns = ['property', 'description']
        return columns, sorted(driver_properties.items())


class ListBaremetalDriverRaidProperty(command.Lister):
    """List a driver's RAID logical disk properties."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ListBaremetalDriverRaidProperty")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'driver',
            metavar='<driver>',
            help='Name of the driver.')
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        raid_props = baremetal_client.driver.raid_logical_disk_properties(
            parsed_args.driver)
        columns = ['property', 'description']
        return columns, sorted(raid_props.items())


class PassthruCallBaremetalDriver(command.ShowOne):
    """Call a vendor passthru method for a driver."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".PassthruCallBaremetalDriver")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'driver',
            metavar='<driver>',
            help=_('Name of the driver.')
        )
        parser.add_argument(
            'method',
            metavar='<method>',
            help=_("Vendor passthru method to be called.")
        )
        parser.add_argument(
            '--arg',
            metavar='<key=value>',
            action='append',
            help=_("Argument to pass to the passthru method (repeat option "
                   "to specify multiple arguments).")
        )
        parser.add_argument(
            '--http-method',
            dest='http_method',
            metavar='<http-method>',
            choices=v1_utils.HTTP_METHODS,
            default='POST',
            help=_("The HTTP method to use in the passthru request. One of "
                   "%s. Defaults to 'POST'.") %
                 oscutils.format_list(v1_utils.HTTP_METHODS)
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        arguments = utils.key_value_pairs_to_dict(parsed_args.arg)
        response = (baremetal_client.driver.
                    vendor_passthru(parsed_args.driver,
                                    parsed_args.method,
                                    http_method=parsed_args.http_method,
                                    args=arguments))

        return self.dict2columns(response)


class PassthruListBaremetalDriver(command.Lister):
    """List available vendor passthru methods for a driver."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".PassthruListBaremetalDriver")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'driver',
            metavar='<driver>',
            help=_('Name of the driver.'))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        columns = res_fields.VENDOR_PASSTHRU_METHOD_RESOURCE.fields

        methods = baremetal_client.driver.get_vendor_passthru_methods(
            parsed_args.driver)

        params: list[dict[str, object]] = []
        for method, response in methods.items():
            response['name'] = method
            http_methods = ', '.join(response['http_methods'])
            response['http_methods'] = http_methods
            params.append(response)

        return (columns,
                (oscutils.get_dict_properties(s, columns) for s in params))


class ShowBaremetalDriver(command.ShowOne):
    """Show information about a driver."""

    log: logging.Logger = logging.getLogger(
        __name__ + ".ShowBaremetalDriver")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'driver',
            metavar='<driver>',
            help=_('Name of the driver.'))
        parser.add_argument(
            '--fields',
            nargs='+',
            dest='fields',
            metavar='<field>',
            action='append',
            default=[],
            choices=res_fields.DRIVER_DETAILED_RESOURCE.fields,
            help=_("One or more node fields. Only these fields will be "
                   "fetched from the server."))
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace,
    ) -> tuple[Sequence[str], Iterable[Any]]:
        self.log.debug("take_action(%s)", parsed_args)
        manager = self.app.client_manager
        baremetal_client = manager.baremetal

        fields: list[str] | None
        fields = (
            list(itertools.chain.from_iterable(parsed_args.fields))
            or None)
        driver = baremetal_client.driver.get(parsed_args.driver,
                                             fields=fields)._info
        driver.pop("links", None)
        driver.pop("properties", None)
        # For list-type properties, show the values as comma separated
        # strings. It's easier to read.
        driver = utils.convert_list_props_to_comma_separated(driver)
        return cast(
            tuple[Sequence[str], Iterable[Any]],
            tuple(zip(*sorted(driver.items()))),
        )



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from http import client as http_client
import json
import time
from typing import Any
from unittest import mock

from keystoneauth1 import exceptions as kexc

from ironicclient.common import filecache
from ironicclient.common import http
from ironicclient import exc
from ironicclient.tests.unit import utils


DEFAULT_TIMEOUT = 600

DEFAULT_HOST = 'localhost'
DEFAULT_PORT = '1234'


def _get_error_body(
    faultstring: str | None = None,
    debuginfo: str | None = None,
    description: str | None = None,
) -> str:
    if description:
        raw_error_body = json.dumps({'description': description})
    else:
        raw_error_body = json.dumps(
            {'faultstring': faultstring, 'debuginfo': debuginfo}
        )
    body = {'error_message': raw_error_body}
    return json.dumps(body)


def _session_client(**kwargs: Any) -> http.SessionClient:
    return http.SessionClient(os_ironic_api_version='1.6',
                              api_version_select_state='default',
                              max_retries=5,
                              retry_interval=2,
                              auth=None,
                              interface='publicURL',
                              service_type='baremetal',
                              region_name='',
                              endpoint_override='http://%s:%s' % (
                                  DEFAULT_HOST, DEFAULT_PORT),
                              **kwargs)


class VersionNegotiationMixinTest(utils.BaseTestCase):

    def setUp(self) -> None:
        super(VersionNegotiationMixinTest, self).setUp()
        self.test_object = http.VersionNegotiationMixin()
        self.test_object.os_ironic_api_version = '1.6'
        self.test_object.api_version_select_state = 'default'
        self.test_object.endpoint_override = "http://localhost:1234"
        self.mock_mcu = mock.MagicMock()
        self.test_object._make_connection_url = self.mock_mcu
        self.response = utils.FakeResponse(
            {}, status=http_client.NOT_ACCEPTABLE)
        self.test_object.get_server = mock.MagicMock(
            return_value=('localhost', '1234'))

    def test__generic_parse_version_headers_has_headers(self) -> None:
        response = {'X-OpenStack-Ironic-API-Minimum-Version': '1.1',
                    'X-OpenStack-Ironic-API-Maximum-Version': '1.6',
                    }
        expected = ('1.1', '1.6')
        result = self.test_object._generic_parse_version_headers(response.get)
        self.assertEqual(expected, result)

    def test__generic_parse_version_headers_missing_headers(self) -> None:
        response = {}
        expected = (None, None)
        result = self.test_object._generic_parse_version_headers(response.get)
        self.assertEqual(expected, result)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    def test_negotiate_version_bad_state(
        self, mock_save_data: mock.MagicMock,
    ) -> None:
        # Test if bad api_version_select_state value
        self.test_object.api_version_select_state = 'word of the day: augur'
        self.assertRaises(
            RuntimeError,
            self.test_object.negotiate_version,
            None, None)
        self.assertEqual(0, mock_save_data.call_count)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_older(
        self,
        mock_pvh: mock.MagicMock,
        mock_save_data: mock.MagicMock,
    ) -> None:
        # Test newer client and older server
        latest_ver = '1.5'
        mock_pvh.return_value = ('1.1', latest_ver)
        mock_conn = mock.MagicMock()
        result = self.test_object.negotiate_version(mock_conn, self.response)
        self.assertEqual(latest_ver, result)
        self.assertEqual(1, mock_pvh.call_count)
        host, port = http.get_server(self.test_object.endpoint_override)
        mock_save_data.assert_called_once_with(host=host, port=port,
                                               data=latest_ver)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_newer(
        self,
        mock_pvh: mock.MagicMock,
        mock_save_data: mock.MagicMock,
    ) -> None:
        # Test newer server and older client
        mock_pvh.return_value = ('1.1', '1.10')
        mock_conn = mock.MagicMock()
        result = self.test_object.negotiate_version(mock_conn, self.response)
        self.assertEqual('1.6', result)
        self.assertEqual(1, mock_pvh.call_count)
        mock_save_data.assert_called_once_with(host=DEFAULT_HOST,
                                               port=DEFAULT_PORT,
                                               data='1.6')

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_make_simple_request',
                       autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_no_version_on_error(
            self,
            mock_pvh: mock.MagicMock,
            mock_msr: mock.MagicMock,
            mock_save_data: mock.MagicMock,
    ) -> None:
        # Test older Ironic version which errored with no version number and
        # have to retry with simple get
        mock_pvh.side_effect = iter([(None, None), ('1.1', '1.2')])
        mock_conn = mock.MagicMock()
        result = self.test_object.negotiate_version(mock_conn, self.response)
        self.assertEqual('1.2', result)
        self.assertTrue(mock_msr.called)
        self.assertEqual(2, mock_pvh.call_count)
        self.assertEqual(1, mock_save_data.call_count)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_explicit_too_high(
        self,
        mock_pvh: mock.MagicMock,
        mock_save_data: mock.MagicMock,
    ) -> None:
        # requested version is not supported because it is too large
        mock_pvh.return_value = ('1.1', '1.6')
        mock_conn = mock.MagicMock()
        self.test_object.api_version_select_state = 'user'
        self.test_object.os_ironic_api_version = '99.99'
        self.assertRaises(
            exc.UnsupportedVersion,
            self.test_object.negotiate_version,
            mock_conn, self.response)
        self.assertEqual(1, mock_pvh.call_count)
        self.assertEqual(0, mock_save_data.call_count)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_explicit_not_supported(
        self,
        mock_pvh: mock.MagicMock,
        mock_save_data: mock.MagicMock,
    ) -> None:
        # requested version is supported by the server but the server returned
        # 406 because the requested operation is not supported with the
        # requested version
        mock_pvh.return_value = ('1.1', '1.6')
        mock_conn = mock.MagicMock()
        self.test_object.api_version_select_state = 'negotiated'
        self.test_object.os_ironic_api_version = '1.5'
        self.assertRaises(
            exc.UnsupportedVersion,
            self.test_object.negotiate_version,
            mock_conn, self.response)
        self.assertEqual(1, mock_pvh.call_count)
        self.assertEqual(0, mock_save_data.call_count)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_strict_version_comparison(
        self,
        mock_pvh: mock.MagicMock,
        mock_save_data: mock.MagicMock,
    ) -> None:
        # Test version comparison with StrictVersion
        max_ver = '1.10'
        mock_pvh.return_value = ('1.2', max_ver)
        mock_conn = mock.MagicMock()
        self.test_object.os_ironic_api_version = '1.10'
        result = self.test_object.negotiate_version(mock_conn, self.response)
        self.assertEqual(max_ver, result)
        self.assertEqual(1, mock_pvh.call_count)
        host, port = http.get_server(self.test_object.endpoint_override)
        mock_save_data.assert_called_once_with(host=host, port=port,
                                               data=max_ver)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_make_simple_request',
                       autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_user_latest(
            self,
            mock_pvh: mock.MagicMock,
            mock_msr: mock.MagicMock,
            mock_save_data: mock.MagicMock,
    ) -> None:
        # For the test to work we need to have the mock reply
        # with a version range that exceeds the client version
        # otherwise we are not confirming that the client is
        # clamping the version correctly
        large_value = http.LAST_KNOWN_API_VERSION + 99
        # have to retry with simple get
        mock_pvh.side_effect = iter([
            (None, None),
            ('1.1', f"1.{large_value}")
        ])
        mock_conn = mock.MagicMock()
        self.test_object.api_version_select_state = 'user'
        self.test_object.os_ironic_api_version = 'latest'
        result = self.test_object.negotiate_version(mock_conn, None)
        self.assertEqual(http.LATEST_VERSION, result)
        self.assertEqual('negotiated',
                         self.test_object.api_version_select_state)
        self.assertEqual(http.LATEST_VERSION,
                         self.test_object.os_ironic_api_version)

        self.assertTrue(mock_msr.called)
        self.assertEqual(2, mock_pvh.call_count)
        self.assertEqual(1, mock_save_data.call_count)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_make_simple_request',
                       autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_user_list(
            self,
            mock_pvh: mock.MagicMock,
            mock_msr: mock.MagicMock,
            mock_save_data: mock.MagicMock,
    ) -> None:
        # have to retry with simple get
        mock_pvh.side_effect = [(None, None), ('1.1', '1.26')]
        mock_conn = mock.MagicMock()
        self.test_object.api_version_select_state = 'user'
        self.test_object.os_ironic_api_version = ['1.1', '1.6', '1.25',
                                                  '1.26', '1.27', '1.28',
                                                  '1.30']
        result = self.test_object.negotiate_version(mock_conn, self.response)
        self.assertEqual('1.26', result)
        self.assertEqual('negotiated',
                         self.test_object.api_version_select_state)
        self.assertEqual('1.26',
                         self.test_object.os_ironic_api_version)

        self.assertTrue(mock_msr.called)
        self.assertEqual(2, mock_pvh.call_count)
        self.assertEqual(1, mock_save_data.call_count)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_make_simple_request',
                       autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_user_list_fails_nomatch(
            self,
            mock_pvh: mock.MagicMock,
            mock_msr: mock.MagicMock,
            mock_save_data: mock.MagicMock,
    ) -> None:
        # have to retry with simple get
        mock_pvh.side_effect = iter([(None, None), ('1.2', '1.26')])
        mock_conn = mock.MagicMock()
        self.test_object.api_version_select_state = 'user'
        self.test_object.os_ironic_api_version = ['1.39', '1.1']
        self.assertRaises(
            exc.UnsupportedVersion,
            self.test_object.negotiate_version,
            mock_conn, self.response)
        self.assertEqual('user',
                         self.test_object.api_version_select_state)
        self.assertEqual(['1.39', '1.1'],
                         self.test_object.os_ironic_api_version)
        self.assertEqual(2, mock_pvh.call_count)
        self.assertEqual(0, mock_save_data.call_count)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_make_simple_request',
                       autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_user_list_single_value(
            self,
            mock_pvh: mock.MagicMock,
            mock_msr: mock.MagicMock,
            mock_save_data: mock.MagicMock,
    ) -> None:
        # have to retry with simple get
        mock_pvh.side_effect = iter([(None, None), ('1.1', '1.26')])
        mock_conn = mock.MagicMock()
        self.test_object.api_version_select_state = 'user'
        # NOTE(TheJulia): Lets test this value explicitly because the
        # minor number is actually the same.
        self.test_object.os_ironic_api_version = ['1.01']
        result = self.test_object.negotiate_version(mock_conn, None)
        self.assertEqual('1.1', result)
        self.assertEqual('negotiated',
                         self.test_object.api_version_select_state)
        self.assertEqual('1.1',
                         self.test_object.os_ironic_api_version)
        self.assertTrue(mock_msr.called)
        self.assertEqual(2, mock_pvh.call_count)
        self.assertEqual(1, mock_save_data.call_count)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_make_simple_request',
                       autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_server_user_list_fails_latest(
            self,
            mock_pvh: mock.MagicMock,
            mock_msr: mock.MagicMock,
            mock_save_data: mock.MagicMock,
    ) -> None:
        # have to retry with simple get
        mock_pvh.side_effect = iter([(None, None), ('1.1', '1.2')])
        mock_conn = mock.MagicMock()
        self.test_object.api_version_select_state = 'user'
        self.test_object.os_ironic_api_version = ['1.01', 'latest']
        self.assertRaises(
            ValueError,
            self.test_object.negotiate_version,
            mock_conn, self.response)
        self.assertEqual('user',
                         self.test_object.api_version_select_state)
        self.assertEqual(['1.01', 'latest'],
                         self.test_object.os_ironic_api_version)
        self.assertEqual(2, mock_pvh.call_count)
        self.assertEqual(0, mock_save_data.call_count)

    @mock.patch.object(filecache, 'save_data', autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_make_simple_request',
                       autospec=True)
    @mock.patch.object(http.VersionNegotiationMixin, '_parse_version_headers',
                       autospec=True)
    def test_negotiate_version_explicit_version_request(
            self,
            mock_pvh: mock.MagicMock,
            mock_msr: mock.MagicMock,
            mock_save_data: mock.MagicMock,
    ) -> None:
        mock_pvh.side_effect = iter([(None, None), ('1.1', '1.99')])
        mock_conn = mock.MagicMock()
        self.test_object.api_version_select_state = 'negotiated'
        self.test_object.os_ironic_api_version = '1.30'
        req_header = {'X-OpenStack-Ironic-API-Version': '1.29'}
        response = utils.FakeResponse(
            {}, status=http_client.NOT_ACCEPTABLE,
            request_headers=req_header)
        self.assertRaisesRegex(exc.UnsupportedVersion,
                               ".*is not supported by the server.*",
                               self.test_object.negotiate_version,
                               mock_conn, response)
        self.assertTrue(mock_msr.called)
        self.assertEqual(2, mock_pvh.call_count)
        self.assertFalse(mock_save_data.called)

    def test_get_server(self) -> None:
        host = 'ironic-host'
        port = '6385'
        endpoint_override = 'http://%s:%s/ironic/v1/' % (host, port)
        self.assertEqual((host, port), http.get_server(endpoint_override))


class SessionClientTest(utils.BaseTestCase):
    def test_server_exception_empty_body(self) -> None:
        error_body = _get_error_body()

        fake_session = utils.mockSession({'Content-Type': 'application/json'},
                                         error_body,
                                         http_client.INTERNAL_SERVER_ERROR)

        client = _session_client(session=fake_session)

        self.assertRaises(exc.InternalServerError,
                          client.json_request,
                          'GET', '/v1/resources')

    def test_server_exception_description_only(self) -> None:
        error_msg = 'test error msg'
        error_body = _get_error_body(description=error_msg)
        fake_session = utils.mockSession(
            {'Content-Type': 'application/json'},
            error_body, status_code=http_client.BAD_REQUEST)
        client = _session_client(session=fake_session)

        self.assertRaisesRegex(exc.BadRequest, 'test error msg',
                               client.json_request,
                               'GET', '/v1/resources')

    def test__parse_version_headers(self) -> None:
        # Test parsing of version headers from SessionClient
        fake_session = utils.mockSession(
            {'X-OpenStack-Ironic-API-Minimum-Version': '1.1',
             'X-OpenStack-Ironic-API-Maximum-Version': '1.6',
             'content-type': 'text/plain',
             },
            None,
            http_client.HTTP_VERSION_NOT_SUPPORTED)
        expected_result = ('1.1', '1.6')
        client = _session_client(session=fake_session)
        result = client._parse_version_headers(fake_session.request())
        self.assertEqual(expected_result, result)

    def test_make_simple_request(self) -> None:
        session = utils.mockSession({})

        client = _session_client(session=session)
        res = client._make_simple_request(session, 'GET', 'url')

        session.request.assert_called_once_with(
            'url', 'GET', raise_exc=False,
            endpoint_filter={
                'interface': 'publicURL',
                'service_type': 'baremetal',
                'region_name': ''
            },
            endpoint_override='http://localhost:1234',
            user_agent=http.USER_AGENT)
        self.assertEqual(res, session.request.return_value)

    @mock.patch.object(http.SessionClient, 'get_endpoint', autospec=True)
    def test_endpoint_not_found(
        self, mock_get_endpoint: mock.MagicMock,
    ) -> None:
        mock_get_endpoint.return_value = None

        self.assertRaises(exc.EndpointNotFound, _session_client,
                          session=utils.mockSession({}))

    def test_json_request(self) -> None:
        session = utils.mockSession({}, status_code=200)
        req_id = "req-7b081d28-8272-45f4-9cf6-89649c1c7a1a"
        client = _session_client(
            session=session, additional_headers={"foo": "bar"},
            global_request_id=req_id)
        client.json_request('GET', 'url')

        session.request.assert_called_once_with(
            'url', 'GET', raise_exc=False, auth=None,
            headers={
                "foo": "bar",
                "X-OpenStack-Request-ID": req_id,
                "Content-Type": "application/json",
                "Accept": "application/json",
                "X-OpenStack-Ironic-API-Version": "1.6"
            },
            endpoint_filter={
                'interface': 'publicURL',
                'service_type': 'baremetal',
                'region_name': ''
            },
            endpoint_override='http://localhost:1234',
            user_agent=http.USER_AGENT
        )


@mock.patch.object(time, 'sleep', lambda *_: None)
class RetriesTestCase(utils.BaseTestCase):

    def test_session_retry(self) -> None:
        error_body = _get_error_body()

        fake_resp = utils.mockSessionResponse(
            {'Content-Type': 'application/json'},
            error_body,
            http_client.CONFLICT)
        ok_resp = utils.mockSessionResponse(
            {'Content-Type': 'application/json'},
            b"OK",
            http_client.OK)
        fake_session = utils.mockSession({})
        fake_session.request.side_effect = iter((fake_resp, ok_resp))

        client = _session_client(session=fake_session)
        client.json_request('GET', '/v1/resources')
        self.assertEqual(2, fake_session.request.call_count)

    def test_session_retry_503(self) -> None:
        error_body = _get_error_body()

        fake_resp = utils.mockSessionResponse(
            {'Content-Type': 'application/json'},
            error_body,
            http_client.SERVICE_UNAVAILABLE)
        ok_resp = utils.mockSessionResponse(
            {'Content-Type': 'application/json'},
            b"OK",
            http_client.OK)
        fake_session = utils.mockSession({})
        fake_session.request.side_effect = iter((fake_resp, ok_resp))

        client = _session_client(session=fake_session)
        client.json_request('GET', '/v1/resources')
        self.assertEqual(2, fake_session.request.call_count)

    def test_session_retry_connection_refused(self) -> None:
        ok_resp = utils.mockSessionResponse(
            {'Content-Type': 'application/json'},
            b"OK",
            http_client.OK)
        fake_session = utils.mockSession({})
        fake_session.request.side_effect = iter((exc.ConnectionRefused(),
                                                 ok_resp))

        client = _session_client(session=fake_session)
        client.json_request('GET', '/v1/resources')
        self.assertEqual(2, fake_session.request.call_count)

    def test_session_retry_retriable_connection_failure(
        self,
    ) -> None:
        ok_resp = utils.mockSessionResponse(
            {'Content-Type': 'application/json'},
            b"OK",
            http_client.OK)
        fake_session = utils.mockSession({})
        fake_session.request.side_effect = iter(
            (kexc.RetriableConnectionFailure(), ok_resp))

        client = _session_client(session=fake_session)
        client.json_request('GET', '/v1/resources')
        self.assertEqual(2, fake_session.request.call_count)

    def test_session_retry_fail(self) -> None:
        error_body = _get_error_body()

        fake_resp = utils.mockSessionResponse(
            {'Content-Type': 'application/json'},
            error_body,
            http_client.CONFLICT)
        fake_session = utils.mockSession({})
        fake_session.request.return_value = fake_resp

        client = _session_client(session=fake_session)

        self.assertRaises(exc.Conflict, client.json_request,
                          'GET', '/v1/resources')
        self.assertEqual(http.DEFAULT_MAX_RETRIES + 1,
                         fake_session.request.call_count)

    def test_session_max_retries_none(self) -> None:
        error_body = _get_error_body()

        fake_resp = utils.mockSessionResponse(
            {'Content-Type': 'application/json'},
            error_body,
            http_client.CONFLICT)
        fake_session = utils.mockSession({})
        fake_session.request.return_value = fake_resp

        client = _session_client(session=fake_session)
        client.conflict_max_retries = None

        self.assertRaises(exc.Conflict, client.json_request,
                          'GET', '/v1/resources')
        self.assertEqual(http.DEFAULT_MAX_RETRIES + 1,
                         fake_session.request.call_count)

    def test_session_change_max_retries(self) -> None:
        error_body = _get_error_body()

        fake_resp = utils.mockSessionResponse(
            {'Content-Type': 'application/json'},
            error_body,
            http_client.CONFLICT)
        fake_session = utils.mockSession({})
        fake_session.request.return_value = fake_resp

        client = _session_client(session=fake_session)
        client.conflict_max_retries = http.DEFAULT_MAX_RETRIES + 1

        self.assertRaises(exc.Conflict, client.json_request,
                          'GET', '/v1/resources')
        self.assertEqual(http.DEFAULT_MAX_RETRIES + 2,
                         fake_session.request.call_count)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import testtools
from testtools.matchers import HasLength

from ironicclient.tests.unit import utils
from ironicclient.v1 import conductor


CONDUCTOR1 = {'hostname': 'compute1.localdomain',
              'conductor_group': 'alpha-team',
              'alive': True,
              'drivers': ['ipmitool', 'fake-hardware'],
              }
CONDUCTOR2 = {'hostname': 'compute2.localdomain',
              'conductor_group': 'alpha-team',
              'alive': True,
              'drivers': ['ipmitool', 'fake-hardware'],
              }

fake_responses = {
    '/v1/conductors':
    {
        'GET': (
            {},
            {"conductors": [CONDUCTOR1, CONDUCTOR2]}
        ),
    },
    '/v1/conductors/?detail=True':
    {
        'GET': (
            {},
            {"conductors": [CONDUCTOR1, CONDUCTOR2]}
        ),
    },
    '/v1/conductors/?fields=hostname,alive':
    {
        'GET': (
            {},
            {"conductors": [CONDUCTOR1]}
        ),
    },
    '/v1/conductors/%s' % CONDUCTOR1['hostname']:
    {
        'GET': (
            {},
            CONDUCTOR1,
        ),
    },
    '/v1/conductors/%s?fields=hostname,alive' % CONDUCTOR1['hostname']:
    {
        'GET': (
            {},
            CONDUCTOR1,
        ),
    },
}

fake_responses_pagination = {
    '/v1/conductors':
    {
        'GET': (
            {},
            {"conductors": [CONDUCTOR1],
             "next": "http://127.0.0.1:6385/v1/conductors/?limit=1"}
        ),
    },
    '/v1/conductors/?limit=1':
    {
        'GET': (
            {},
            {"conductors": [CONDUCTOR2]}
        ),
    },
    '/v1/conductors/?marker=%s' % CONDUCTOR1['hostname']:
    {
        'GET': (
            {},
            {"conductors": [CONDUCTOR2]}
        ),
    },
}

fake_responses_sorting = {
    '/v1/conductors/?sort_key=updated_at':
    {
        'GET': (
            {},
            {"conductors": [CONDUCTOR2, CONDUCTOR1]}
        ),
    },
    '/v1/conductors/?sort_dir=desc':
    {
        'GET': (
            {},
            {"conductors": [CONDUCTOR2, CONDUCTOR1]}
        ),
    },
}


class ConductorManagerTest(testtools.TestCase):

    def setUp(self) -> None:
        super(ConductorManagerTest, self).setUp()
        self.api = utils.FakeAPI(fake_responses)
        self.mgr = conductor.ConductorManager(self.api)

    def test_conductor_list(self) -> None:
        conductors = self.mgr.list()
        expect = [
            ('GET', '/v1/conductors', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(conductors))

    def test_conductor_list_detail(self) -> None:
        conductors = self.mgr.list(detail=True)
        expect = [
            ('GET', '/v1/conductors/?detail=True', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(conductors))

    def test_conductor_list_limit(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = conductor.ConductorManager(self.api)
        conductors = self.mgr.list(limit=1)
        expect = [
            ('GET', '/v1/conductors/?limit=1', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(conductors, HasLength(1))

    def test_conductor_list_marker(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = conductor.ConductorManager(self.api)
        conductors = self.mgr.list(marker=CONDUCTOR1['hostname'])
        expect = [
            ('GET', '/v1/conductors/?marker=%s' % CONDUCTOR1['hostname'],
             {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertThat(conductors, HasLength(1))

    def test_conductor_list_pagination_no_limit(self) -> None:
        self.api = utils.FakeAPI(fake_responses_pagination)
        self.mgr = conductor.ConductorManager(self.api)
        conductors = self.mgr.list(limit=0)
        expect = [
            ('GET', '/v1/conductors', {}, None),
            ('GET', '/v1/conductors/?limit=1', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(conductors))

    def test_conductor_list_sort_key(self) -> None:
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = conductor.ConductorManager(self.api)
        conductors = self.mgr.list(sort_key='updated_at')
        expect = [
            ('GET', '/v1/conductors/?sort_key=updated_at', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(conductors))

    def test_conductor_list_sort_dir(self) -> None:
        self.api = utils.FakeAPI(fake_responses_sorting)
        self.mgr = conductor.ConductorManager(self.api)
        conductors = self.mgr.list(sort_dir='desc')
        expect = [
            ('GET', '/v1/conductors/?sort_dir=desc', {}, None)
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(2, len(conductors))

    def test_conductor_list_fields(self) -> None:
        conductors = self.mgr.list(fields=['hostname', 'alive'])
        expect = [
            ('GET', '/v1/conductors/?fields=hostname,alive', {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(1, len(conductors))

    def test_conductor_show(self) -> None:
        conductor = self.mgr.get(CONDUCTOR1['hostname'])
        expect = [
            ('GET', '/v1/conductors/%s' % CONDUCTOR1['hostname'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(CONDUCTOR1['hostname'], conductor.hostname)

    def test_conductor_show_fields(self) -> None:
        conductor = self.mgr.get(CONDUCTOR1['hostname'],
                                 fields=['hostname', 'alive'])
        expect = [
            ('GET', '/v1/conductors/%s?fields=hostname,alive' %
             CONDUCTOR1['hostname'], {}, None),
        ]
        self.assertEqual(expect, self.api.calls)
        self.assertEqual(CONDUCTOR1['hostname'], conductor.hostname)
